# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common test-helper code involving django usage."""

import abc
import asyncio
import contextlib
import copy
import datetime as dt
import http.client
import inspect
import io
import itertools
import json
import traceback
import unittest
from collections.abc import Callable, Generator, Iterator, Sequence
from collections.abc import Collection as AbcCollection
from operator import itemgetter
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Protocol,
    TYPE_CHECKING,
    cast,
    override,
    runtime_checkable,
)
from unittest import mock
from unittest.util import safe_repr

import django.template
import django.test
import lxml
import requests
from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.conf import settings
from django.core.signals import request_finished
from django.db import (
    OperationalError,
    close_old_connections,
    connection,
    connections,
    transaction,
)
from django.db.models import Model, QuerySet
from django.http import HttpRequest
from django.http.response import HttpResponseBase
from django.template import engines
from django.test.client import JSON_CONTENT_TYPE_RE
from django.urls import set_urlconf
from django.utils import timezone
from rest_framework import status

from debusine.db.context import ContextConsistencyError, context
from debusine.db.models import (
    Collection,
    CollectionItemMatchConstraint,
    Scope,
    Token,
    User,
    WorkRequest,
    Workspace,
)
from debusine.db.models.permissions import (
    Allow,
    PermissionCheckPredicate,
    get_resource_scope,
    get_resource_workspace,
    resolve_roles_list,
)
from debusine.db.models.work_requests import WorkRequestQuerySet
from debusine.db.permissioncontext import PermissionContext, PermissionUser
from debusine.db.playground.scenarios import Scenario
from debusine.db.tasks import DBTask, DBWorkerTask
from debusine.tasks.models import (
    BaseDynamicTaskData,
    BaseTaskData,
    BaseTaskDataModel,
)
from debusine.test.playground import Playground

# Relative import needed to break an import loop
from .base import TestCase as BaseTestCase  # noqa: ABS101, TID252

if TYPE_CHECKING:
    from django.test.client import _MonkeyPatchedWSGIResponse

    TestMixinBase = unittest.TestCase
else:
    TestMixinBase = object

#: Type alias for response objects returned by the Django test client
type TestResponseType = "_MonkeyPatchedWSGIResponse"
#: Type alias for any response type
type AnyResponseType = HttpResponseBase | "_MonkeyPatchedWSGIResponse"


@contextlib.contextmanager
def preserve_db_task_registry() -> Generator[None]:
    """Make task registry changes ephemeral."""
    db_orig = DBTask._sub_tasks
    DBTask._sub_tasks = copy.deepcopy(DBTask._sub_tasks)
    try:
        yield
    finally:
        DBTask._sub_tasks = db_orig


class BaseTestWorkerTask[TD: BaseTaskData](
    DBWorkerTask[TD, BaseDynamicTaskData], metaclass=abc.ABCMeta
):
    """Base class for simple worker tasks used in test code."""

    @override
    def get_task_configuration_subject_context(
        self,
    ) -> tuple[str | None, str | None]:
        raise NotImplementedError()

    @override
    def compute_dynamic_data(self) -> BaseDynamicTaskData:
        raise NotImplementedError()


class PermissionOverride[M: Model](abc.ABC):
    """Base class for configurable permission overrides for tests."""

    @abc.abstractmethod
    def check(self, obj: M, user: PermissionUser | None) -> bool:
        """Test the predicate on an object."""

    @abc.abstractmethod
    def filter(
        self, queryset: QuerySet[M, M], user: PermissionUser | None
    ) -> QuerySet[M, M]:
        """Filter a QuerySet with the predicate."""


class AllowAll[M: Model](PermissionOverride[M]):
    """Permission override allowing all."""

    @override
    def check(self, obj: M, user: PermissionUser | None) -> bool:  # noqa: ARG002, U100
        return True

    @override
    def filter(
        self,
        queryset: QuerySet[M, M],
        user: PermissionUser | None,  # noqa: ARG002, U100
    ) -> QuerySet[M, M]:
        return queryset


class DenyAll[M: Model](PermissionOverride[M]):
    """Permission override denying all."""

    @override
    def check(self, obj: M, user: PermissionUser | None) -> bool:  # noqa: ARG002, U100
        return False

    @override
    def filter(
        self,
        queryset: QuerySet[M, M],
        user: PermissionUser | None,  # noqa: ARG002, U100
    ) -> QuerySet[M, M]:
        return queryset.none()


class ListFilter[M: Model](PermissionOverride[M]):
    """Permission override that includes or excludes specific items."""

    def __init__(
        self, include: list[M] | None = None, exclude: list[M] | None = None
    ) -> None:
        """Store filter arguments."""
        super().__init__()
        self._include = include
        self._exclude = exclude

    @override
    def check(self, obj: M, user: PermissionUser | None) -> bool:  # noqa: ARG002, U100
        if self._include is not None:
            return obj in self._include
        if self._exclude is not None:
            return obj not in self._exclude
        return True

    @override
    def filter(
        self,
        queryset: QuerySet[M, M],
        user: PermissionUser | None,  # noqa: ARG002, U100
    ) -> QuerySet[M, M]:
        res = queryset
        if self._include is not None:
            res = res.filter(pk__in=[x.pk for x in self._include])
        if self._exclude is not None:
            res = res.exclude(pk__in=[x.pk for x in self._exclude])
        return res


def set_context_for_resource(resource: Model, user: PermissionUser) -> None:
    """Set the current context to fit the resource."""
    scope = get_resource_scope(resource)
    assert scope is not None
    context.set_permission_context_scoped(
        scope,
        (
            PermissionContext.for_ui_user(user)
            if user.is_authenticated
            else PermissionContext.for_ui_anonymous()
        ),
    )
    if workspace := get_resource_workspace(resource):
        workspace.set_current()


@contextlib.contextmanager
def override_permission[M: Model](
    model_cls: type[M],
    pred: str,
    override_cls: type[PermissionOverride[M]],
    **kwargs: Any,
) -> Generator[None]:
    """
    Override a permission.

    Example usage::

        with override_permission(
            Workspace, "can_display", AllowAll
        ):
            ...

    Or also::

        @override_permission(
            Workspace, "can_display", AllowAll
        )
        def test_method(self) -> None:
            ...
    """
    override = override_cls(**kwargs)

    # Model.objects is a manager instance generated by from_queryset, and we
    # need a bit of creativity to mock its methods
    def on_manager(user: PermissionUser | None) -> QuerySet[M, M]:
        return override.filter(getattr(model_cls, "objects").all(), user)

    manager = getattr(model_cls, "objects")
    queryset_cls = manager.get_queryset().__class__
    error_template = getattr(getattr(model_cls, pred), "error_template")
    with (
        mock.patch.object(
            model_cls, pred, autospec=True, side_effect=override.check
        ) as check_patch,
        mock.patch.object(
            manager,
            pred,
            autospec=True,
            side_effect=on_manager,
        ),
        mock.patch.object(
            queryset_cls,
            pred,
            autospec=True,
            side_effect=override.filter,
        ),
    ):
        setattr(check_patch, "error_template", error_template)
        yield


@runtime_checkable
class JSONResponseProtocol(Protocol):
    """A Django test client response with a monkey-patched json() method."""

    def json(self) -> Any:
        """Return the body of the response, parsed as JSON."""


class BaseDjangoTestCase(django.test.TransactionTestCase, BaseTestCase):
    """
    Django-specific Debusine test methods.

    This augments debusine.test.TestCase with django-specific assert statements
    and factory functions.

    This is the common base class for TestCase and TransactionTestCase, for
    tests that do depend on Django code.
    """

    @override
    @classmethod
    def get_precious_globals(cls) -> dict[str, Any]:
        """Collect global values that tests should not change."""
        res = super().get_precious_globals()
        res["templates_0_dirs"] = settings.TEMPLATES[0]["DIRS"]
        return res

    @override
    @classmethod
    def setUpClass(cls) -> None:
        """Run class-wide test setup in a temporary context."""
        with context.local():
            context.reset()
            super().setUpClass()

    @override
    def setUp(self) -> None:
        """Ensure test methods do not change the global scope."""
        super().setUp()
        self.enterContext(context.local())
        context.reset()
        set_urlconf(None)

    @override
    def tearDown(self) -> None:
        # Some tests change Django's URL configuration.  Django resets this
        # at the start and end of each request, but that doesn't necessarily
        # happen in tests.  If a test that uses reverse() runs with a
        # non-default URL configuration, then it may spuriously fail.  Reset
        # the URL configuration at the end of each test to make sure that
        # doesn't happen.
        set_urlconf(None)
        super().tearDown()

    @contextlib.contextmanager
    def custom_template(self, path: str | Path) -> Generator[Path]:
        """
        Install a custom template, removing it at context manager exit.

        It generates a path pointing to the custom template file, but nothing
        is written: you can use the path you get to write out the template
        contents.
        """
        # Clear template caches before writing the new template and after
        # removing it: if we are changing an included template, Django doesn't
        # seem able to notice that the cache needs invalidating
        path = Path(path)
        template_path = Path(settings.DEBUSINE_TEMPLATE_DIRECTORY) / path
        template_path.parent.mkdir(exist_ok=True, parents=True)
        try:
            self.clear_template_cache()
            yield template_path
        finally:
            template_path.unlink(missing_ok=True)
            self.clear_template_cache()

    def clear_template_cache(self) -> None:
        """Clear the Django template cache."""
        from django.template.backends.django import DjangoTemplates
        from django.template.loaders.base import Loader

        for engine in django.template.engines.all():
            if not isinstance(engine, DjangoTemplates):
                continue
            for loader in engine.engine.template_loaders:
                if not isinstance(loader, Loader):
                    continue
                loader.reset()

    @contextlib.contextmanager
    def ephemeral_savepoint(self) -> Generator[Any]:
        """
        Run the code inside a savepoint, rolled back at the end.

        Generates the transaction identifier.

        Rollback only happens if the code succeeded: a runaway exception is
        left to the care of the transaction inside which we are running.
        """
        sid = transaction.savepoint()
        yield sid
        transaction.savepoint_rollback(sid)

    @contextlib.contextmanager
    def separate_connection(self, alias: str) -> Generator[None]:
        """Temporarily open a separate database connection."""
        connections[alias] = temp_connection = connection.copy(alias)
        try:
            yield
        finally:
            del connections[alias]
            temp_connection.close()

    def assertResponseHTML(
        self,
        response: TestResponseType,
        status_code: int = status.HTTP_200_OK,
        content_type: str = "text/html; charset=utf-8",
        dump_on_error: bool = False,
    ) -> lxml.objectify.ObjectifiedElement:
        """
        Check that the response is valid HTML.

        :param response: the Django response to check
        :param status_code: expected status code
        :param content_type: expected content type
        :returns: the parsed HTML contents
        """
        self.assertEqual(response.status_code, status_code)
        self.assertEqual(response.headers["Content-Type"], content_type)
        return self.assertHTMLValid(
            response.content, dump_on_error=dump_on_error
        )

    def assertResponse400(
        self, response: "_MonkeyPatchedWSGIResponse", error: str
    ) -> None:
        """Assert that response is Http400 and contents is in response."""
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.context["error"], error)

    def format_api_response(self, response: AnyResponseType) -> str:
        """
        Format an API response.

        The result is intended to be compact and useful in error messages and
        exception annotations.
        """
        status_description = http.client.responses.get(
            response.status_code, 'unknown'
        )
        content_type = response.get("content-type")
        res = f"{response.status_code} {status_description} ({content_type}): "

        if hasattr(response, "content"):
            assert content_type is not None
            content = response.content.decode(response.charset)
            if content_type == "application/problem+json":
                # Format a ProblemResponse
                decoded = json.loads(content)
                title = decoded.get("title", "(no title)")
                res += f"{title=!r}"
                if detail := decoded.get("detail"):
                    res += f" {detail=!r}"
                if validation_errors := decoded.get("validation_errors"):
                    res += f" {validation_errors=!r}"
                if original_exception := getattr(
                    response, "_original_exception", None
                ):
                    res += " original exception follows:\n"
                    exception_formatted = traceback.format_exception(
                        original_exception
                    )
                    res += "".join(exception_formatted)
            elif JSON_CONTENT_TYPE_RE.match(content_type):
                content = repr(json.loads(content))
                if len(content) > 200:
                    content = content[:200] + "…"
                res += f"content={content}"
            else:
                if len(content) > 200:
                    content = content[:200] + "…"
                res += f"{content=!r}"
        else:
            res += "content not available"
        return res

    def assertAPIResponseOk(
        self, response: AnyResponseType, status_code: int = status.HTTP_200_OK
    ) -> Any:
        """
        Ensure an API response succeeded with the given status code.

        :returns: the decoded JSON contents.
        """
        try:
            self.assertEqual(response.status_code, status_code)

            if status_code == status.HTTP_204_NO_CONTENT:
                assert (
                    not hasattr(response, "content") or response.content == b""
                )
                return None
            content_type = response.get("Content-Type")
            if content_type is None:
                self.fail("response has no content type")
            if not JSON_CONTENT_TYPE_RE.match(content_type):
                self.fail(
                    f"content type {content_type!r} is not a JSON response"
                )

            assert hasattr(response, "content")
            return json.loads(response.content.decode(response.charset))
        except Exception as exc:
            exc.add_note(self.format_api_response(response))
            raise

    def assertResponseProblem(
        self,
        response: HttpResponseBase,
        title: str,
        detail_pattern: str | None = None,
        validation_errors_pattern: str | None = None,
        status_code: int = requests.codes.bad_request,
    ) -> None:
        """
        Assert that response is a valid application/problem+json.

        Assert that the content_type is application/problem+json and the
        title exist and matches title.

        :param response: response that it is asserting
        :param status_code: assert response.status_code == status_code
        :param title: exact match with response.data["title"]
        :param detail_pattern: if not None: assertRegex with
           response.data["detail"]. If None checks that response.data does not
           contain "detail".
        :param validation_errors_pattern: if not None: assertRegex with
           response.data["validation_errors"]. If None checks that
           response.data does not contain "validation_errors".
        """
        self.assertEqual(
            response.status_code,
            status_code,
            f"response status {response.status_code} != {status_code}",
        )

        content_type = response.headers["Content-Type"]
        self.assertEqual(
            content_type,
            "application/problem+json",
            f'content_type "{content_type}" != "application/problem+json"',
        )

        assert isinstance(response, JSONResponseProtocol)
        data = response.json()

        self.assertIn("title", data, '"title" not found in response')

        response_title = data["title"]
        self.assertEqual(
            response_title, title, f'title "{response_title}" != "{title}"'
        )

        if detail_pattern is not None:
            self.assertIn("detail", data, '"detail" not found in response')

            response_detail = str(data["detail"])
            self.assertRegex(
                response_detail,
                detail_pattern,
                f'Detail regexp "{detail_pattern}" did not '
                f'match "{response_detail}"',
            )
        else:
            self.assertNotIn("detail", data, '"detail" is in the response')

        if validation_errors_pattern is not None:
            self.assertIn(
                "validation_errors",
                data,
                '"validation_errors" not found in response',
            )
            self.assertIsInstance(data["validation_errors"], dict)

            response_validation_errors = str(data["validation_errors"])
            self.assertRegex(
                response_validation_errors,
                validation_errors_pattern,
                f'Validation errors regexp "{validation_errors_pattern}" did '
                f'not match "{response_validation_errors}"',
            )
        else:
            self.assertNotIn(
                "validation_errors",
                data,
                '"validation_errors" is in the response',
            )

    def assertRolePredicateResult(
        self,
        obj: Model,
        name: str,
        role: str,
        *,
        pc: PermissionContext,
        expected_when_has_role: bool = True,
    ) -> None:
        """
        Check the result of a predicate that only depends on the user's role.

        :param obj: an instance of the model. The actual value does not matter,
          as testing is done via mocking.
        :param name: predicate name
        :param role: role that the predicate should check
        :param pc: the PermissionContext to use for checking
        :param expected_when_has_role: the expected result of the predicate
          when the user has the role
        """
        model = obj.__class__
        model_name = model.__name__

        # Checks that model.name and model.objects.name exist
        perm_check = getattr(obj, name, None)
        if perm_check is None:
            self.fail(f"{model_name} has no @permission_check {name}")
        if not hasattr(perm_check, "work_request") or not hasattr(
            perm_check, "anonymous"
        ):
            self.fail(f"{model_name}.{name} is not marked as @permission_check")

        # Test the check predicate
        with mock.patch.object(
            obj, "has_role", autospec=True, return_value=False
        ) as has_role:
            self.assertFalse(perm_check(pc))
        has_role.assert_called_with(pc, role)

        with mock.patch.object(
            obj, "has_role", autospec=True, return_value=True
        ) as has_role:
            self.assertEqual(perm_check(pc), expected_when_has_role)
        has_role.assert_called_with(pc, role)

        # Test the filter predicate on manager
        manager = getattr(model, "objects")
        manager_name = f"{model_name}.objects"
        perm_filter_manager = getattr(manager, name, None)
        if perm_filter_manager is None:
            self.fail(f"{manager_name} has no @permission_filter {name}")
        if not hasattr(perm_filter_manager, "work_request") or not hasattr(
            perm_filter_manager, "anonymous"
        ):
            self.fail(
                f"{manager_name}.{name} is not marked as @permission_filter"
            )

        queryset_class = manager.get_queryset().__class__

        with mock.patch.object(
            queryset_class, "with_role", return_value=manager.filter(pk=obj.pk)
        ) as with_role:
            filtered = perm_filter_manager(pc)
            self.assertIsInstance(filtered, queryset_class)
            if expected_when_has_role:
                self.assertQuerySetEqual(filtered, [obj])
            else:
                self.assertQuerySetEqual(filtered, [])
        with_role.assert_called_with(pc, role)

        with mock.patch.object(
            queryset_class, "with_role", return_value=manager.none()
        ) as with_role:
            filtered = perm_filter_manager(pc)
            self.assertIsInstance(filtered, queryset_class)
            self.assertQuerySetEqual(filtered, [])
        with_role.assert_called_with(pc, role)

        # Test the filter predicate on QuerySet
        queryset = getattr(model, "objects").all()
        queryset_name = f"QuerySet[{model_name}]"

        perm_filter_queryset = getattr(queryset, name, None)
        if perm_filter_queryset is None:
            self.fail(f"{queryset_name} has no @permission_filter {name}")
        if not hasattr(perm_filter_queryset, "work_request") or not hasattr(
            perm_filter_queryset, "anonymous"
        ):
            self.fail(
                f"{queryset_name}.{name} is not marked as @permission_filter"
            )

        with mock.patch.object(
            queryset, "with_role", return_value=queryset.filter(pk=obj.pk)
        ) as with_role:
            filtered = perm_filter_queryset(pc)
            self.assertIsInstance(filtered, queryset_class)
            if expected_when_has_role:
                self.assertQuerySetEqual(filtered, [obj])
            else:
                self.assertQuerySetEqual(filtered, [])
        with_role.assert_called_with(pc, role)

        with mock.patch.object(
            queryset, "with_role", return_value=queryset.none()
        ) as with_role:
            filtered = perm_filter_queryset(pc)
            self.assertIsInstance(filtered, queryset_class)
            self.assertQuerySetEqual(filtered, [])
        with_role.assert_called_with(pc, role)

    def assertRolePredicate(
        self,
        obj: Model,
        name: str,
        role: str,
        *,
        work_request: Allow = Allow.NEVER,
        anonymous: Allow = Allow.NEVER,
    ) -> None:
        """
        Test a predicate only checks if the user has a role.

        :param obj: an instance of the model. The actual value does not matter,
                    as testing is done via mocking.
        :param name: predicate name
        :param role: role that the predicate should check
        """
        user = User()
        pc = PermissionContext.for_ui_user(user)
        self.assertRolePredicateResult(
            obj, name, role, pc=pc, expected_when_has_role=True
        )

        model = obj.__class__

        # Check that the work_request= and anonymous= arguments match

        perm_check = getattr(obj, name)
        self.assertEqual(perm_check.work_request, work_request)
        self.assertEqual(perm_check.anonymous, anonymous)

        manager = getattr(model, "objects")
        perm_filter_manager = getattr(manager, name)
        self.assertEqual(perm_filter_manager.work_request, work_request)
        self.assertEqual(perm_filter_manager.anonymous, anonymous)

        queryset = getattr(model, "objects").all()
        perm_filter_queryset = getattr(queryset, name)
        self.assertEqual(perm_filter_queryset.work_request, work_request)
        self.assertEqual(perm_filter_queryset.anonymous, anonymous)

    def assertPermissionPredicate(
        self,
        pred: Callable[[PermissionContext], bool],
        pc: PermissionContext,
        expected_pass: bool,
    ) -> None:
        """
        Check the result of a permission predicate.

        :param pred: the permission predicate, as the check method on a
                     resource object
        :param user: the user to check
        :param expected_pass: True if the predicate is expected to pass
        """
        resource: Model = pred.__self__  # type: ignore[attr-defined]
        self.assertEqual(pred(pc), expected_pass)
        filtered = getattr(
            resource.__class__.objects,  # type: ignore[attr-defined]
            pred.__name__,
        )(pc)
        if expected_pass:
            self.assertIn(resource, list(filtered))
        else:
            self.assertNotIn(resource, list(filtered))

    def _assertPermission_context(
        self, user: PermissionUser | None = None, token: Token | None = None
    ) -> PermissionContext:
        if token is None:
            assert user is not None
            if user.is_authenticated:
                return PermissionContext.for_ui_user(user)
            else:
                return PermissionContext.for_ui_anonymous()

        match token.token_type:
            case Token.Types.WORKER_ACTIVATION:
                return PermissionContext.for_worker_activation(token)
            case Token.Types.WORKER:
                return PermissionContext.for_worker(token)
            case Token.Types.WORK_REQUEST:
                return PermissionContext.for_work_request(token=token)
            case Token.Types.USER:
                return PermissionContext.for_api_user(token)
            case Token.Types.EXTERNAL_USER:
                return PermissionContext.for_api_external_user(token)
            case _ as unreachable:
                # Enforced by database constraint.
                raise AssertionError(f"Unexpected token type: {unreachable}")

    @context.local()
    def _test_check_predicate[M: Model](
        self,
        pc: PermissionContext,
        obj: M,
        name: str,
        expected: bool,
    ) -> list[str]:
        """Test a single check predicate on an object."""
        errors: list[str] = []
        predicate = getattr(obj, name)

        # Test on an empty context
        context.reset()

        if predicate(pc) is not expected:
            errors.append(
                f"{name} is {not expected} for {pc} on {obj} with empty context"
            )

        # Try with scope set in context, to check shortcut test code
        if (scope := get_resource_scope(obj)) is None:
            return errors

        try:
            context.set_permission_context_scoped(scope, pc)
            if predicate(context.pc) is not expected:
                errors.append(
                    f"{name} is {not expected} for {pc} on {obj}"
                    " with scope set in context"
                )

            if (workspace := get_resource_workspace(obj)) is None:
                return errors

            # Try with also workspace set in context, to check shortcut test
            # code
            workspace.set_current()
            if predicate(context.pc) is not expected:
                errors.append(
                    f"{name} is {not expected} for {pc} on {obj}"
                    " with scope and workspace set in context"
                )
        except ContextConsistencyError:
            pass

        return errors

    def _check_predicate_filter_output(
        self,
        prefix: str,
        filtered: list[str],
        expected: list[str],
        errors: list[str],
    ) -> None:
        if filtered == expected:
            return
        error = prefix
        if filtered:
            error += ", ".join(filtered)
        else:
            error += "nothing"
        error += " instead of "
        if expected:
            error += ", ".join(expected)
        else:
            error += "nothing"
        errors.append(error)

    @context.local()
    def assertPermission[M: Model](
        self,
        name: str,
        users: PermissionUser | None | AbcCollection[PermissionUser | None],
        allowed: M | AbcCollection[M] = (),
        denied: M | AbcCollection[M] = (),
        token: Token | None = None,
    ) -> None:
        """
        Test a permission predicate.

        For each user given, test that the model objects in ``allowed`` allow
        the named permission, and those in ``denied`` do not.

        Both check and filter predicates are tested. Note that, because the
        filter predicate is tested, ``allowed`` must list all objects in the
        database that test true for the permission.
        """
        if not isinstance(users, AbcCollection):
            users = (users,)
        if not isinstance(allowed, AbcCollection):
            allowed = (allowed,)
        if not isinstance(denied, AbcCollection):
            denied = (denied,)

        errors: list[str] = []

        # Ensure permission is not granted when using worker activation tokens
        worker_activation_token = Token(
            token_type=Token.Types.WORKER_ACTIVATION, enabled=True
        )
        worker_activation_pc = self._assertPermission_context(
            None, worker_activation_token
        )
        for obj in itertools.chain(allowed, denied):
            errors += self._test_check_predicate(
                worker_activation_pc, obj, name, False
            )

        # Ensure permission is not granted when using worker tokens
        worker_token = Token(token_type=Token.Types.WORKER, enabled=True)
        worker_pc = self._assertPermission_context(None, worker_token)
        for obj in itertools.chain(allowed, denied):
            errors += self._test_check_predicate(worker_pc, obj, name, False)

        # We cannot use subTest because the reported stack trace would stop at
        # subTest, and we would lose information on where assertPermission is
        # called
        for user in users:
            pc = self._assertPermission_context(user, token)
            for obj in allowed:
                errors += self._test_check_predicate(pc, obj, name, True)

            for obj in denied:
                errors += self._test_check_predicate(pc, obj, name, False)

            # Pick the model class from the first element of allow or
            # denied
            model_cls = next(itertools.chain(allowed, denied)).__class__
            queryset = getattr(model_cls, "objects")
            filter_predicate = getattr(queryset, name)

            # Check that there are no allowed resources that weren't in the
            # allowed list
            filtered = sorted(str(obj) for obj in filter_predicate(pc))
            expected = sorted(str(obj) for obj in allowed)
            self._check_predicate_filter_output(
                f"{name} for {user} selects ", filtered, expected, errors
            )

            # Check that there are no allowed resources that weren't in the
            # allowed list
            filtered = sorted(
                str(obj)
                for obj in queryset.exclude(pk__in=filter_predicate(pc))
            )
            expected = sorted(str(obj) for obj in denied)
            self._check_predicate_filter_output(
                f"{name} for {user} denies ", filtered, expected, errors
            )

        if errors:
            with io.StringIO() as buf:
                print("Predicate permission mismatch:", file=buf)
                for error in errors:
                    print(f"* {error}", file=buf)
                self.fail(buf.getvalue())

    @contextlib.contextmanager
    def assert_model_count_unchanged(
        self, model_cls: type[Model]
    ) -> Generator[None]:
        """Context manager to assert count of rows in a model is unchanged."""
        initial_count = getattr(model_cls, "objects").count()
        try:
            yield
        finally:
            self.assertEqual(
                getattr(model_cls, "objects").count(), initial_count
            )

    @contextlib.contextmanager
    def template_dir(self) -> Generator[Path]:
        """Prepend a temporary directory to the template search paths."""
        template_dir = self.create_temporary_directory()
        template_settings = copy.deepcopy(settings.TEMPLATES)
        cast(list[str], template_settings[0]["DIRS"]).insert(
            0, template_dir.as_posix()
        )
        with django.test.override_settings(TEMPLATES=template_settings):
            yield template_dir

    def assert_collection_has_constraints(
        self,
        collection: Collection,
        expected_constraints: list[dict[str, Any]],
    ) -> None:
        """Assert that a collection has these constraints."""
        constraints = [
            dict(constraint)
            for constraint in CollectionItemMatchConstraint.objects.filter(
                collection=collection
            ).values()
        ]
        for constraint in constraints:
            del constraint["id"]
        self.assertEqual(
            sorted(constraints, key=itemgetter("collection_item_id", "key")),
            sorted(
                [
                    {"collection_id": collection.id, **constraint}
                    for constraint in expected_constraints
                ],
                key=itemgetter("collection_item_id", "key"),
            ),
        )

    def close_streaming_response(self, response: HttpResponseBase) -> None:
        """Close a streaming response without reading the rest of its data."""
        # See https://code.djangoproject.com/ticket/35618
        #
        # django.test.client.ClientHandler does this when closing a response in
        # the non-streaming case, but I cannot find a standard way to close a
        # streaming response without invalidating the database connection held
        # by the tests, which will cause further tests to fail.
        #
        # This is a copy of what django.test.client.ClientHandler does in the
        # non-streaming case
        request_finished.disconnect(close_old_connections)
        response.close()  # will fire request_finished
        request_finished.connect(close_old_connections)

    def render_string(
        self,
        template_code: str,
        request: HttpRequest,
        **context: Any,
    ) -> str:
        """Render a template from a string."""
        template = engines["django"].from_string(template_code)
        return template.render(context, request=request)

    @contextlib.contextmanager
    def freeze_time(self) -> Iterator[dt.datetime]:
        """Freeze django.utils.timezone.now() timestamps."""
        frozen_time = timezone.now()
        with mock.patch("django.utils.timezone.now", return_value=frozen_time):
            yield frozen_time


class PlaygroundTestCase(BaseDjangoTestCase):
    """
    Playground-specific test case setup.

    Class annotations are scanned in test case setup for Scenario instances,
    which are automatically instantiated and set in the class.

    Scenario instances can be annotated with a dict, which will provide extra
    arguments passed to playground.scenario(). This can be used, for example,
    to call ``set_current``::

      scenario: Annotated[ClassVar[DefaultContext], {"set_current": True}]
    """

    #: set to False if you do not need a playground in your test case
    playground_needed: bool = True

    #: set to False if you need a playground with a LocalFileStore instead of
    #: an in-memory version
    playground_memory_file_store: bool = True

    @classmethod
    def create_playground(cls) -> Playground:
        """
        Create the playground object.

        This method can be overridden by subclasses to create a playground with
        a different configuration
        """
        return Playground(memory_file_store=cls.playground_memory_file_store)

    @classmethod
    def list_playground_scenarios(
        cls,
    ) -> Generator[tuple[str, Scenario]]:
        """
        List the scenarios configured in the test case.

        Generate tuples like ``(name, scenario)``.
        """
        for name, value in inspect.getmembers(cls):
            if not isinstance(value, Scenario):
                continue
            yield name, value

    def _build_permission_when_role_testlist(
        self,
        predicate: Callable[[PermissionContext], bool],
        roles: str | AbcCollection[str] = (),
        *,
        scope_roles: str | AbcCollection[str] = (),
        workspace_roles: str | AbcCollection[str] = (),
    ) -> list[tuple[Model, str]]:
        """Build a TODO list for testing the predicate with the given roles."""
        assert isinstance(predicate, PermissionCheckPredicate)
        resource = predicate.__self__
        roles = resolve_roles_list(resource, roles)
        scope_roles = resolve_roles_list(Scope, scope_roles)
        workspace_roles = resolve_roles_list(Workspace, workspace_roles)

        tests: list[tuple[Model, str]] = []
        if scope_roles:
            scope = get_resource_scope(resource)
            assert scope is not None
            tests.extend((scope, role) for role in scope_roles)
        if workspace_roles:
            workspace = get_resource_workspace(resource)
            assert workspace is not None
            tests.extend((workspace, role) for role in workspace_roles)
        tests.extend((resource, role) for role in roles)

        return tests

    def assertNoPermissionWhenRole(  # noqa: C901
        self,
        predicate: Callable[[PermissionContext], bool],
        pc: PermissionContext,
        roles: str | AbcCollection[str] = (),
        *,
        scope_roles: str | AbcCollection[str] = (),
        workspace_roles: str | AbcCollection[str] = (),
    ) -> None:
        """
        Check that assigning roles does not make the predicate come true.

        Each role is tested in isolation: the test is repeated for each given
        role.
        """
        assert isinstance(predicate, PermissionCheckPredicate)
        errors: list[str] = []
        resource: Model = predicate.__self__
        pred_name = f"{predicate.__func__.__name__!r} on {str(resource)!r}"
        filter_predicate = getattr(
            getattr(resource.__class__, "objects"), predicate.__func__.__name__
        )
        for assignee, role in self._build_permission_when_role_testlist(
            predicate,
            roles,
            scope_roles=scope_roles,
            workspace_roles=workspace_roles,
        ):
            with context.local(), self.ephemeral_savepoint():
                context.reset()

                if predicate(pc):
                    errors.append(f"{pred_name} is true even without {role}")
                if filter_predicate(pc).filter(pk=resource.pk).exists():
                    errors.append(
                        f"{pred_name} passes filtering even without {role}"
                    )

                # Ignoring type until we find a way to type self.playground to
                # accept either a member or a classvar, depending on subclasses
                self.playground.create_group_role(  # type: ignore[attr-defined]
                    assignee,
                    role,
                    users=[pc.user],
                    name="assertNoPermissionWhenRole",
                )

                if predicate(pc):
                    errors.append(f"{pred_name} is true when given {role}")
                if filter_predicate(pc).filter(pk=resource.pk).exists():
                    errors.append(
                        f"{pred_name} passes filtering when given {role}"
                    )

                # Once more with context set, to test shortcutting
                set_context_for_resource(assignee, pc.user)

                if predicate(pc):
                    errors.append(
                        f"{pred_name} is true when given {role} in context"
                    )
                if filter_predicate(pc).filter(pk=resource.pk).exists():
                    errors.append(
                        f"{pred_name} passes filtering when given {role}"
                        " in context"
                    )

        self.assertEqual(errors, [])

    def assertPermissionWhenRole(  # noqa: C901
        self,
        predicate: Callable[[PermissionContext], bool],
        pc: PermissionContext,
        roles: str | AbcCollection[str] = (),
        scope_roles: str | AbcCollection[str] = (),
        workspace_roles: str | AbcCollection[str] = (),
    ) -> None:
        """
        Check that assigning roles makes the predicate come true.

        Each role is tested in isolation: the test is repeated for each given
        role.
        """
        assert isinstance(predicate, PermissionCheckPredicate)
        errors: list[str] = []
        resource = predicate.__self__
        pred_name = f"{predicate.__func__.__name__!r} on {str(resource)!r}"
        filter_predicate = getattr(
            resource.__class__.objects, predicate.__func__.__name__
        )
        for assignee, role in self._build_permission_when_role_testlist(
            predicate,
            roles,
            scope_roles=scope_roles,
            workspace_roles=workspace_roles,
        ):
            with context.local(), self.ephemeral_savepoint():
                context.reset()

                if predicate(pc):
                    errors.append(f"{pred_name} is true even without {role}")
                if filter_predicate(pc).filter(pk=resource.pk).exists():
                    errors.append(
                        f"{pred_name} passes filtering even without {role}"
                    )

                # Ignoring type until we find a way to type self.playground to
                # accept either a member or a classvar, depending on subclasses
                self.playground.create_group_role(  # type: ignore[attr-defined]
                    assignee,
                    role,
                    users=[pc.user],
                    name="assertPermissionWhenRole",
                )

                if not predicate(pc):
                    errors.append(f"{pred_name} is false even with {role}")
                if not filter_predicate(pc).filter(pk=resource.pk).exists():
                    errors.append(
                        f"{pred_name} does not pass filtering even with {role}"
                    )

                # Once more with context set, to test shortcutting
                set_context_for_resource(assignee, pc.user)

                if not predicate(pc):
                    errors.append(
                        f"{pred_name} is false even with {role} in context"
                    )
                if not filter_predicate(pc).filter(pk=resource.pk).exists():
                    errors.append(
                        f"{pred_name} does not pass filtering even with {role}"
                        " in context"
                    )

        self.assertEqual(errors, [])


class TransactionTestCase(
    PlaygroundTestCase,
    django.test.TransactionTestCase,
):
    """Debusine-specific extensions to django's TransactionTestCase."""

    playground: Playground

    @override
    def setUp(self) -> None:
        """Create a playground for TransactionTestCase."""
        super().setUp()
        if self.playground_needed:
            # TransactionTestCase does not support setUpTestData, so we need a
            # playground instance per test method
            self.playground = self.enterContext(self.create_playground())
            # Instantiate playground scenarios
            for name, scenario in self.__class__.list_playground_scenarios():
                # Make a copy of the blank scenario in the class
                method_scenario = copy.deepcopy(scenario)
                # Instantiate a different one for each test method
                self.playground.build_scenario(
                    method_scenario, scenario_name=name, set_current=True
                )
                setattr(self, name, method_scenario)

    def assert_work_requests_locked(
        self, *, locked: Sequence[WorkRequest], unlocked: Sequence[WorkRequest]
    ) -> None:
        """Check that the given work requests are locked/unlocked for update."""
        with self.separate_connection(alias := "test-locks"):
            qs: QuerySet[WorkRequest] = WorkRequestQuerySet(
                WorkRequest, using=alias
            )
            for wr in locked:
                with (
                    transaction.atomic(using=alias),
                    self.assertRaisesRegex(
                        OperationalError, "could not obtain lock"
                    ),
                ):
                    # .exists() is enough here, since the query set contains
                    # at most one row.
                    qs.select_for_update(nowait=True, no_key=True).filter(
                        id=wr.id
                    ).exists()
            for wr in unlocked:
                with transaction.atomic(using=alias):
                    # .exists() is enough here, since the query set contains
                    # at most one row.
                    qs.select_for_update(nowait=True, no_key=True).filter(
                        id=wr.id
                    ).exists()


class TestCase(
    PlaygroundTestCase,
    django.test.TestCase,
):
    """Debusine-specific extensions to django's TestCase."""

    playground: ClassVar[Playground]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Create a playground to setup test data."""
        super().setUpTestData()
        if cls.playground_needed:
            cls.playground = cls.enterClassContext(cls.create_playground())
            for name, scenario in cls.list_playground_scenarios():
                # set_current is delayed to the setUp method
                cls.playground.build_scenario(
                    scenario, scenario_name=name, set_current=False
                )

    @override
    def setUp(self) -> None:
        """Create a playground for TransactionTestCase."""
        super().setUp()
        if self.playground_needed:
            for name, scenario in self.playground.scenarios.items():
                # Replace the class version of the scenario with its deepcopied
                # version
                setattr(self, name, scenario)
                # BaseDjangoTestCase.setUp calls context.reset, so we need to
                # call set_current here instead of in setUpTestData
                if scenario.needs_set_current:
                    scenario.set_current()

    def assert_work_request_event_reactions(
        self,
        work_request: WorkRequest,
        *,
        on_creation: Sequence[BaseTaskDataModel] | None = None,
        on_unblock: Sequence[BaseTaskDataModel] | None = None,
        on_assignment: Sequence[BaseTaskDataModel] | None = None,
        on_success: Sequence[BaseTaskDataModel] | None = None,
        on_failure: Sequence[BaseTaskDataModel] | None = None,
    ) -> None:
        """
        Assert that a work request has the given event reactions.

        The order of the items in each list of event reactions does not
        matter.  Leaving parameters unset or None is equivalent to passing
        the empty list, meaning that they are required to be empty.
        """
        self.assertCountEqual(
            work_request.event_reactions.on_creation or [],
            on_creation or [],
        )
        self.assertCountEqual(
            work_request.event_reactions.on_unblock or [],
            on_unblock or [],
        )
        self.assertCountEqual(
            work_request.event_reactions.on_assignment or [],
            on_assignment or [],
        )
        self.assertCountEqual(
            work_request.event_reactions.on_success or [],
            on_success or [],
        )
        self.assertCountEqual(
            work_request.event_reactions.on_failure or [],
            on_failure or [],
        )


class ChannelsHelpersMixin(TestMixinBase):
    """
    Channel-related methods to help writing unit tests.

    Provides methods to setup a channel and assert messages or lack of messages.
    """

    @override
    def tearDown(self) -> None:
        """Flush the channel layer."""
        super().tearDown()
        channel_layer = get_channel_layer()
        assert channel_layer is not None
        async_to_sync(channel_layer.flush)()

    async def create_channel(self, group_name: str) -> dict[str, Any]:
        """
        Create a channel and add it to the group named ``group_name``.

        Return dict with layer and name.
        """
        channel_layer = get_channel_layer()
        assert channel_layer is not None
        channel_name = await channel_layer.new_channel()
        await channel_layer.group_add(group_name, channel_name)

        return {"layer": channel_layer, "name": channel_name}

    async def assert_channel_nothing_received(
        self, channel: dict[str, Any]
    ) -> None:
        """Assert that nothing is received in channel."""
        try:
            received = await asyncio.wait_for(
                channel["layer"].receive(channel["name"]), timeout=0.1
            )
        except asyncio.exceptions.TimeoutError:
            pass
        else:
            cast(unittest.TestCase, self).fail(
                f"Expected nothing. Received: '{safe_repr(received)}'"
            )

    async def assert_channel_received(
        self, channel: dict[str, Any], data: dict[str, Any]
    ) -> None:
        """Assert that data is received in channel_layer, channel_name."""
        try:
            received = await asyncio.wait_for(
                channel["layer"].receive(channel["name"]), timeout=0.1
            )
            cast(unittest.TestCase, self).assertEqual(received, data)
        except asyncio.exceptions.TimeoutError:
            cast(unittest.TestCase, self).fail(
                f"Expected '{safe_repr(data)}' received nothing"
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data migration to populate ``Token.(activating_)worker``."""

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps
from django.db.models import OuterRef, Subquery

from debusine.db.models.auth import _TokenTypes


def set_token_type_and_workers(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Set ``Token.worker`` and ``Token.activating_worker`."""
    Token = apps.get_model("db", "Token")
    Worker = apps.get_model("db", "Worker")

    for worker_column, token_column in (
        ("token", "_worker_id"),
        ("activation_token", "_activating_worker_id"),
    ):
        Token.objects.filter(
            id__in=Worker.objects.filter(
                **{f"{worker_column}__isnull": False}
            ).values(worker_column)
        ).update(
            **{
                token_column: Subquery(
                    Worker.objects.filter(
                        **{worker_column: OuterRef("id")}
                    ).values("id")
                )
            }
        )

    Token.objects.filter(_activating_worker_id__isnull=False).update(
        token_type=_TokenTypes.WORKER_ACTIVATION
    )
    Token.objects.filter(_worker_id__isnull=False).update(
        token_type=_TokenTypes.WORKER
    )
    Token.objects.filter(user__isnull=False).update(token_type=_TokenTypes.USER)

    # Delete any remaining bare tokens.  They can't be used for anything any
    # more, and the next database migration will add a constraint forbidding
    # them.
    Token.objects.filter(token_type="").delete()


def clear_token_type_and_workers(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Clear ``Token.worker`` and ``Token.activating_worker`."""
    Token = apps.get_model("db", "Token")
    Token.objects.update(
        token_type="", _worker_id=None, _activating_worker_id=None
    )


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0048_mirror_worker_token_ids"),
    ]

    operations = [
        migrations.RunPython(
            set_token_type_and_workers, clear_token_type_and_workers
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""URLs related to user (e.g. token handling)."""

from django.urls import URLPattern, URLResolver, include, path

import debusine.web.views.auth as views

app_name = "user"

user_urlpatterns: list[URLPattern | URLResolver] = [
    path(
        "token/",
        views.UserTokenListView.as_view(),
        name="token-list",
    ),
    path(
        "token/create/",
        views.UserTokenCreateView.as_view(),
        name="token-create",
    ),
    path(
        "token/<int:pk>/edit/",
        views.UserTokenUpdateView.as_view(),
        name="token-edit",
    ),
    path(
        "token/<int:pk>/delete/",
        views.UserTokenDeleteView.as_view(),
        name="token-delete",
    ),
    path(
        "groups/",
        views.UserGroupListView.as_view(),
        name="groups",
    ),
    path(
        "",
        views.UserDetailView.as_view(),
        name="detail",
    ),
]

urlpatterns = [
    path(
        "<int:pk>/",
        views.UserIDRedirectView.as_view(),
        name="id",
    ),
    path(
        "<str:username>/",
        include(user_urlpatterns),
    ),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for users templatetags."""

from django.urls import reverse

from debusine.db.context import context
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.templatetags.users import user_link


class UserTagsTests(TestCase):
    """Tests for users tags."""

    scenario = scenarios.DefaultScopeUser()

    def test_user_link_can_display(self) -> None:
        user = self.playground.create_user("test")
        user_url = reverse("user:detail", kwargs={"username": user.username})
        self.scenario.set_current()

        self.assertEqual(
            user_link(user), f'<a href="{user_url}">{user.username}</a>'
        )

    def test_user_link_cannot_display(self) -> None:
        user = self.playground.create_user("test")
        context.set_permission_context_scoped(
            self.scenario.scope, PermissionContext.for_ui_anonymous()
        )

        self.assertEqual(user_link(user), user.username)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the base task field resolver."""

import re
from contextlib import ExitStack
from typing import Any, Self, override
from unittest import mock

from debusine.tasks import TaskConfigError
from debusine.tasks.inputs import inputs
from debusine.tasks.inputs.mock_resolver import ConcreteFieldResolver
from debusine.tasks.inputs.resolver import FieldResolver
from debusine.test import TestCase


class ConcreteDataListFieldInput(inputs.DataListFieldInput[int, None]):
    """Concrete version of DataListFieldInput."""

    @override
    def resolve(self, field_resolver: FieldResolver) -> None:
        raise NotImplementedError()


class FieldResolverTests(TestCase):
    """Test :py:class:`FieldResolver`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver = ConcreteFieldResolver()
        self.get_field = self.enterContext(
            mock.patch.object(self.resolver, "get_field")
        )

    def test_get_field_collection(self) -> None:
        field = ConcreteDataListFieldInput(field="name")
        for value in ([], [1], (2, 3), {3, 4}):
            with self.subTest(value=repr(value)):
                self.get_field.return_value = value
                self.assertEqual(
                    self.resolver.get_field_collection(field), value
                )

    def test_get_field_collection_none(self) -> None:
        field = ConcreteDataListFieldInput(field="name")
        self.get_field.return_value = None
        self.assertEqual(self.resolver.get_field_collection(field), [])

    def test_get_field_invalid_item_type(self) -> None:
        field = ConcreteDataListFieldInput(field="name")
        value: Any
        for value in ([None], ["foo"], [[], 1]):
            with self.subTest(value=repr(value)):
                self.get_field.return_value = value
                with self.assertRaisesRegex(
                    TaskConfigError,
                    re.escape(
                        f"Task data field name[0] is {value[0]!r}"
                        f" which is not of type <class 'int'>"
                    ),
                ):
                    self.resolver.get_field_collection(field)


class MockResolve(ExitStack):
    """Mock TaskInput.resolve_to_task to track invocations."""

    def __init__(self) -> None:
        """Initialize an empty callee list."""
        super().__init__()
        self.called: list[str] = []

    def __call__(
        self,
        inp: inputs.SingleInput,
        resolver: FieldResolver,  # noqa: ARG002
    ) -> None:
        """Replace TaskInput.resolve_to_task."""
        self.called.append(inp.field)

    @override
    def __enter__(self) -> Self:
        """Install mock resolution for SingleInput."""
        super().__enter__()
        self.enter_context(
            mock.patch(
                "debusine.tasks.inputs.SingleInput.resolve",
                side_effect=self,
                autospec=True,
            )
        )
        return self


class ResolveInputsTests(TestCase):
    """Test :py:func:`resolve_inputs`."""

    @override
    def setUp(self) -> None:
        super().setUp()

        class _Container(inputs.TaskFieldContainer):
            all = inputs.SingleInput(field="all")
            config = inputs.SingleInput(
                field="config", stage=inputs.Stage.PENDING
            )

        self.task = _Container()
        self.resolver = ConcreteFieldResolver()

    def test_resolve_inputs_default(self) -> None:
        """Calling without stage runs all tasks."""
        with MockResolve() as resolve_to_task:
            self.resolver.resolve_inputs(self.task)
        self.assertEqual(resolve_to_task.called, ["all", "config"])

    def test_resolve_inputs_partial_stage(self) -> None:
        """Calling with a stage runs only on the tasks of that stage."""
        with MockResolve() as resolve_to_task:
            self.resolver.resolve_inputs(self.task, stage=inputs.Stage.PENDING)
        self.assertEqual(resolve_to_task.called, ["config"])

    def test_resolve_inputs_running(self) -> None:
        """Calling with stage RUNNING runs all tasks."""
        with MockResolve() as resolve_to_task:
            self.resolver.resolve_inputs(self.task, stage=inputs.Stage.RUNNING)
        self.assertEqual(resolve_to_task.called, ["all", "config"])

    def test_resolve_inputs_multiple_times(self) -> None:
        with MockResolve() as resolve_to_task:
            self.resolver.resolve_inputs(self.task, stage=inputs.Stage.PENDING)
        self.assertEqual(resolve_to_task.called, ["config"])

        # Previously resolved inputs are re-resolved
        with MockResolve() as resolve_to_task:
            self.resolver.resolve_inputs(self.task, stage=inputs.Stage.RUNNING)
        self.assertEqual(resolve_to_task.called, ["all", "config"])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Template tag library to render artifacts."""

from django import template
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from debusine.artifacts.models import ArtifactCategory
from debusine.db.models import Artifact
from debusine.db.models.artifacts import (
    ARTIFACT_CATEGORY_ICON_NAMES,
    ARTIFACT_CATEGORY_SHORT_NAMES,
)

register = template.Library()


@register.filter
def artifact_icon_name(artifact: Artifact | str) -> str:
    """Return the Bootstrap icon name for this artifact's category."""
    if isinstance(artifact, Artifact):
        category = artifact.category
    else:
        category = artifact
    return ARTIFACT_CATEGORY_ICON_NAMES.get(ArtifactCategory(category), "file")


@register.filter
def artifact_category_label(artifact: Artifact | str) -> str:
    """Return a short label to use to represent an artifact category."""
    if isinstance(artifact, Artifact):
        category = artifact.category
    else:
        category = artifact
    return ARTIFACT_CATEGORY_SHORT_NAMES.get(
        ArtifactCategory(category), "artifact"
    )


@register.simple_tag
def artifact_link(artifact: int | Artifact) -> str:
    """Return a link to the artifact if it exists or "${artifact_id} Deleted."""

    def render(artifact_obj: Artifact) -> str:
        return format_html(
            '<a href="{}">{}</a>',
            artifact_obj.get_absolute_url(),
            artifact_obj.get_label(),
        )

    if isinstance(artifact, Artifact):
        return render(artifact)

    try:
        artifact_obj = Artifact.objects.get(id=artifact)
        return render(artifact_obj)
    except Artifact.DoesNotExist:
        return mark_safe(f"{artifact} (deleted)")


@register.simple_tag
def artifact_links(artifact_ids: list[int]) -> str:
    """Return links to the artifacts if exists or "${artifact_id} Deleted."""
    artifacts_by_id = {
        a.id: a for a in Artifact.objects.filter(id__in=artifact_ids)
    }

    # Some "artifact_ids" might not exist: use the Artifact or the id when
    # calling "artifact_link" to generate the ("deleted" artifacts)
    links = [
        artifact_link(artifacts_by_id.get(artifact_id, artifact_id))
        for artifact_id in artifact_ids
    ]
    return mark_safe(", ".join(links))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utility code for testing views."""

import datetime as dt
import re
from collections.abc import Callable
from typing import Any
from unittest import mock

import lxml.etree
import lxml.objectify
from django.contrib.auth.models import AnonymousUser
from django.http import HttpRequest
from django.template import engines
from django.test import RequestFactory, override_settings
from django.urls import resolve, reverse
from django.utils.formats import date_format as django_date_format
from django.views.generic.base import View
from rest_framework import status

from debusine.db.context import ContextConsistencyError
from debusine.db.models import User, WorkRequest, WorkflowTemplate, Workspace
from debusine.db.permissioncontext import PermissionContext
from debusine.test.django import (
    AllowAll,
    BaseDjangoTestCase,
    DenyAll,
    override_permission,
)
from debusine.web.views.base import Widget


class ViewTestMixin(BaseDjangoTestCase):
    """TestCase functions used to test Debusine views."""

    def make_request(self, url: str) -> HttpRequest:
        """Create a request that can be manipulated before invoking a view."""
        factory = RequestFactory()
        request = factory.get(url)
        request.user = AnonymousUser()
        return request

    def instantiate_view_class[ViewClass: View](
        self,
        view_class: type[ViewClass],
        request_or_url: HttpRequest | str,
        **kwargs: Any,
    ) -> ViewClass:
        """
        Instantiate a View subclass with the given request.

        For convenience, if request is a string it will be passed to
        make_request.
        """
        request: HttpRequest
        if isinstance(request_or_url, str):
            request = self.make_request(request_or_url)
        else:
            request = request_or_url
        view = view_class()
        view.setup(request, **kwargs)
        return view

    @staticmethod
    def normalize_node(node: lxml.objectify.ObjectifiedElement) -> str:
        """Normalize the HTML to ignore spaces and new lines."""

        def remove_new_lines_blanks(s: str) -> str:
            # Multiple spaces to single space
            s = re.sub(r"\s+", " ", s)

            # New lines and trailing spaces are removed
            return s.replace("\n", "").strip()

        root = lxml.etree.fromstring(lxml.etree.tostring(node))
        for element in root.iter():
            if element.text:
                element.text = remove_new_lines_blanks(element.text)
            if element.tail:
                element.tail = remove_new_lines_blanks(element.tail)
        return lxml.etree.tostring(root, method="html", encoding="unicode")

    def assertHTMLContentsEquivalent(
        self,
        node: lxml.objectify.ObjectifiedElement,
        expected: str,
    ) -> None:
        """Ensure that node's HTML is equivalent to expected_html."""
        expected_element = lxml.objectify.fromstring(expected)

        normalized_html_node = self.normalize_node(node)
        normalized_html_expected = self.normalize_node(expected_element)

        self.assertEqual(normalized_html_node, normalized_html_expected)

    def assertWorkRequestRow(
        self, tr: lxml.objectify.ObjectifiedElement, work_request: WorkRequest
    ) -> None:
        """Ensure the row shows the given work request."""
        work_request_url = work_request.get_absolute_url()
        self.assertTextContentEqual(tr.td[0], str(work_request.id))
        self.assertEqual(tr.td[0].a.get("href"), work_request_url)
        self.assertEqual(
            tr.td[1].get("title"),
            django_date_format(work_request.created_at, "DATETIME_FORMAT"),
        )
        self.assertTextContentEqual(tr.td[2], work_request.task_type)
        self.assertTextContentEqual(
            tr.td[3], work_request.get_task().get_label()
        )
        self.assertTextContentEqual(tr.td[4], work_request.status.capitalize())
        self.assertTextContentEqual(tr.td[5], work_request.result.capitalize())

    def workspace_list_table_rows(
        self, tree: lxml.etree._Element
    ) -> list[lxml.objectify.ObjectifiedElement]:
        """Find the workspace list table in the page and return it."""
        table = tree.xpath("//table[@id='workspace-list-table']")
        if not table:
            self.fail("page has no workspace list table")
        return list(table[0].tbody.tr)

    def assertSetsCurrentWorkspace(
        self,
        workspace: Workspace,
        url: str,
        method: str = "get",
        **kwargs: Any,
    ) -> None:
        """
        Check that the view sets the current workspace correctly.

        :param workspace: Workspace that is supposed to get set
        :param url: URL of the view to call
        :param method: HTTP method to use
        :param kwargs: passed to ``client.{method}``
        """

        class Reached(BaseException):
            """Thrown when the target code is reached."""

        with mock.patch(
            "debusine.db.models.Workspace.set_current",
            side_effect=ContextConsistencyError("expected fail"),
        ):
            response = getattr(self.client, method)(url, **kwargs)
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
            self.assertRegex(
                response.context["exception"],
                r"Workspace .+ not found in scope .+, or you are not "
                r"authorized to see it",
            )

        with (
            self.assertRaises(Reached),
            mock.patch(
                "debusine.db.models.Workspace.set_current",
                autospec=True,
                side_effect=Reached(),
            ) as set_current,
        ):
            response = getattr(self.client, method)(url, **kwargs)
        set_current.assert_called_once_with(workspace, check_visibility=True)

    def assertEnforcesPermission(
        self,
        predicate: Callable[[PermissionContext], bool],
        url: str,
        target_name: str,
        method: str = "get",
        error_code: int = status.HTTP_403_FORBIDDEN,
        **kwargs: Any,
    ) -> None:
        """
        Check that the view enforces the given permission.

        :param predicate: predicate to check
        :param url: URL of the view to call
        :param target_name: name (to be used with mock.patch) of the view
                       method that gets called after permissions are checked.
                       If it does not contain any dot, it is intended as a
                       method of the view class used for ``url``
        :param method: HTTP method to use
        :param kwargs: passed to ``client.{method}``

        """

        class Reached(BaseException):
            """Thrown when the target code is reached."""

        if "." not in target_name:
            view_class = getattr(resolve(url).func, "view_class")
            target_name = (
                f"{view_class.__module__}.{view_class.__qualname__}"
                f".{target_name}"
            )

        with (
            mock.patch(target_name, side_effect=Reached()) as target,
            override_permission(
                getattr(predicate, "__self__").__class__,
                predicate.__name__,
                AllowAll,
            ),
            self.assertRaises(Reached),
        ):
            response = getattr(self.client, method)(url, **kwargs)
        target.assert_called()

        with (
            mock.patch(target_name, side_effect=Reached()) as target,
            override_permission(
                getattr(predicate, "__self__").__class__,
                predicate.__name__,
                DenyAll,
            ),
        ):
            response = getattr(self.client, method)(url, **kwargs)
            self.assertEqual(response.status_code, error_code)
        target.assert_not_called()

    def assertNavCommonElements(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        """Check that the common elements in the navbar are present."""
        el = self.assertHasElement(tree, "//a[@id='navbar-homepage-link']")
        self.assertEqual(el.get("href"), reverse("homepage:homepage"))

    def assertNavNoWorkspaces(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        """Check that the page has no workspaces in the navbar."""
        self.assertFalse(tree.xpath("//li[@class='nav-workspaces']"))

    def assertNavHasWorkspaces(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        current: Workspace,
        others: list[Workspace],
    ) -> None:
        """Check that the navbar has the given current and other workspaces."""
        el = self.assertHasElement(tree, "//li[@id='nav-workspaces']")
        if others:
            self.assertEqual(
                el.div.a[0].get("href"), current.get_absolute_url()
            )
            expected = [li.a.get("href") for li in el.div.ul.li]
            actual = [ws.get_absolute_url() for ws in others]
            self.assertEqual(expected, actual)
        else:
            self.assertEqual(el.a.get("href"), current.get_absolute_url())
            self.assertTextContentEqual(el.a, current.name)

    def assertNavNoWorkflows(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        """Check that the page has no collections in the navbar."""
        self.assertFalse(tree.xpath("//li[@class='nav-workflows']"))

    def assertNavHasWorkflows(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        workspace: Workspace,
        workflow_templates: list[WorkflowTemplate],
    ) -> None:
        """Check that the page has the given collections in the navbar."""
        nav = self.assertHasElement(tree, "//li[@id='nav-workflows']")
        if not workflow_templates:
            workflows_a = self.assertHasElement(nav, "a")
            self.assertFalse(nav.xpath("div"))
        else:
            workflows_a = self.assertHasElement(nav, "div/a[@href!='#']")

        list_url = reverse(
            "workspaces:workflows:list", kwargs={"wname": workspace.name}
        )
        self.assertEqual(workflows_a.get("href"), list_url)
        self.assertTextContentEqual(workflows_a, "Workflows")

        if not workflow_templates:
            return

        workflow_list_base_url = reverse(
            "workspaces:workflows:list",
            kwargs={"wname": workspace.name},
        )
        menu = self.assertHasElement(nav, "*/ul")
        for wt, li in zip(workflow_templates, menu.li):
            a = li.a
            self.assertTextContentEqual(a, wt.name)
            url = a.get("href")
            assert url is not None
            self.assertIn(workflow_list_base_url, url)
            paginator = self.client.get(url).context["paginator"]
            self.assertEqual(
                paginator.table.filters["workflow_templates"].value,
                wt.name,
            )

    def assertNavNoCollections(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        """Check that the page has no collections in the navbar."""
        self.assertFalse(tree.xpath("//li[@class='nav-collections']"))

    def assertNavHasCollections(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        workspace: Workspace,
    ) -> None:
        """Check that the page has the given collections in the navbar."""
        el = self.assertHasElement(tree, "//li[@id='nav-collections']")
        self.assertEqual(
            el.div.a[0].get("href"),
            reverse(
                "workspaces:collections:list",
                kwargs={"wname": workspace.name},
            ),
        )
        self.assertTextContentEqual(el.div.a[0], "Collections")

        expected = {
            (c.get_absolute_url(), str(c)) for c in workspace.collections.all()
        }
        actual = set()
        for li in el.div.ul.li:
            actual.add((li.a.get("href"), li.a.text))
        self.assertEqual(expected, actual)

    def assertNavNoUser(self, tree: lxml.objectify.ObjectifiedElement) -> None:
        """Check that the page has no user in the navbar."""
        ul = self.assertHasElement(tree, "//ul[@id='navbar-right']")
        login_url = ul.li.a.get("href")
        assert login_url is not None
        self.assertEqual(login_url.split("?")[0], reverse("login"))

        self.assertFalse(tree.xpath("//ul[@id='navbar-user-actions']"))

    def assertNavHasUser(
        self, tree: lxml.objectify.ObjectifiedElement, user: User
    ) -> None:
        """Check that the page navbar has a user menu for the given user."""
        ul = self.assertHasElement(tree, "//ul[@id='navbar-right']")
        self.assertTextContentEqual(ul.li.a, user.username)

        ul = self.assertHasElement(tree, "//ul[@id='navbar-user-actions']")
        self.assertEqual(
            ul.li[0].a.get("href"),
            reverse("user:detail", kwargs={"username": user.username}),
        )
        self.assertEqual(
            ul.li[1].a.get("href"),
            reverse("user:token-list", kwargs={"username": user.username}),
        )
        self.assertEqual(ul.li[-1].form.get("action"), reverse("logout"))

    def assertNavHasPlumbing(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        """Check that the page navbar has "Plumbing" drop down."""
        a = self.assertHasElement(tree, "//a[@id='navbar-plumbing-actions']")
        self.assertTextContentEqual(a, "Plumbing")

    def assertNavNoPlumbing(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        """Check that the page navbar has no "Plumbing" drop down."""
        self.assertFalse(tree.xpath("//a[@id='navbar-plumbing-actions']"))

    def render_widget(self, widget: Widget) -> str:
        """Render a widget."""
        template = engines["django"].from_string(
            "{% load debusine %}{% widget widget %}"
        )
        setattr(template, "engine", engines["django"])
        # Set DEBUG=True to have render raise exceptions instead of logging
        # them
        with override_settings(DEBUG=True):
            return template.render({"widget": widget})


def html_check_icon(value: bool) -> str:
    """Return HTML for check icon."""
    if value:
        return '<i style="color:green;" class="bi bi-check2"></i>'
    else:
        return '<i style="color:red;" class="bi bi-x"></i>'


def date_format(d: dt.datetime) -> str:
    """Return dt datetime formatted with the Django template format."""
    return django_date_format(d, "DATETIME_FORMAT")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the models."""

from collections.abc import Sequence
from typing import override

from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.utils import IntegrityError

from debusine.db.models import File, FileInStore, FileStore
from debusine.db.models.files import FileStoreJournal
from debusine.db.tests.utils import (
    RunInParallelTransaction,
    _calculate_hash_from_data,
)
from debusine.server.file_backend.external import ExternalDebianSuiteFileBackend
from debusine.server.file_backend.local import LocalFileBackend
from debusine.server.file_backend.memory import MemoryFileBackend
from debusine.test.django import TestCase, TransactionTestCase


class FileManagerTests(TestCase):
    """Unit tests for the ``FileManager`` class."""

    @override
    def setUp(self) -> None:
        """Set up File to be used in the tests."""
        super().setUp()
        self.file_contents = b"test"
        self.file_hash = _calculate_hash_from_data(self.file_contents)

        self.file = self.playground.create_file(self.file_contents)

    def test_get_or_create_not_created(self) -> None:
        """
        File.objects.get_or_create returns created=False.

        File.objects.get_or_create() tried to create a file with the same
        file_hash and size as one already existing.
        """
        fileobj, created = File.objects.get_or_create(
            hash_digest=self.file_hash, size=len(self.file_contents)
        )

        self.assertEqual(self.file, fileobj)
        self.assertFalse(created)

    def test_get_or_create_created(self) -> None:
        """
        File.objects.get_or_create returns created=True.

        File.objects.get_or_create() created a new File (file_hash and size
        did not exist).
        """
        file_contents = self.file_contents + b"-new-contents"
        hash_digest = _calculate_hash_from_data(file_contents)
        size = len(file_contents)

        fileobj, created = File.objects.get_or_create(
            hash_digest=hash_digest, size=size
        )

        self.assertIsInstance(fileobj, File)
        self.assertEqual(fileobj.hash_digest, hash_digest)
        self.assertEqual(fileobj.size, size)
        self.assertTrue(created)


class FileTests(TestCase):
    """Tests for the File class."""

    @override
    def setUp(self) -> None:
        """Set up File to be used in the tests."""
        super().setUp()
        self.file_contents = b"test"
        self.file_hash = _calculate_hash_from_data(self.file_contents)

        self.file = self.playground.create_file(self.file_contents)

    def test_sha256(self) -> None:
        """File.sha256 have the expected hash."""
        self.assertEqual(self.file.sha256, self.file_hash)
        self.assertEqual(self.file.sha256.hex(), self.file_hash.hex())

    def test_hash_digest_getter(self) -> None:
        """File.hash_digest return the expected hash digest."""
        self.assertIsInstance(self.file.hash_digest, bytes)
        self.assertEqual(self.file.hash_digest, self.file_hash)
        self.assertEqual(self.file.hash_digest.hex(), self.file_hash.hex())

    def test_hash_digest_setter(self) -> None:
        """File.hash_digest setter sets the hash digest."""
        the_hash = _calculate_hash_from_data(b"some data")
        self.file.hash_digest = the_hash

        self.file.save()
        self.file.refresh_from_db()

        self.assertEqual(self.file.hash_digest, the_hash)

    def test_calculate_hash(self) -> None:
        """calculate_hash returns the expected hash digest."""
        file_contents = b"testing"
        local_file = self.create_temporary_file(contents=file_contents)
        expected_hash = _calculate_hash_from_data(file_contents)

        self.assertEqual(
            self.file.calculate_hash(
                local_file,
            ),
            expected_hash,
        )

    def test_str(self) -> None:
        """__str__() return sha256 and size."""
        self.assertEqual(
            str(self.file),
            f"id: {self.file.id} "
            f"sha256: {self.file_hash.hex()} "
            f"size: {len(self.file_contents)}",
        )

    def test_unique_constraint_hash_size(self) -> None:
        """File with same hash_digest and size cannot be created."""
        with self.assertRaisesRegex(
            IntegrityError, "db_file_unique_sha256_size"
        ):
            File.objects.create(
                **{
                    File.current_hash_algorithm: self.file.hash_digest,
                    "size": self.file.size,
                }
            )

    def test_hash_can_be_duplicated(self) -> None:
        """File with the same hash_digest and different size can be created."""
        file_existing = File.objects.earliest("id")
        file_new, _ = File.objects.get_or_create(
            hash_digest=file_existing.hash_digest, size=file_existing.size + 10
        )
        self.assertIsNotNone(file_new.id)
        self.assertNotEqual(file_new.id, file_existing.id)

    def test_constraint_hash_digest_not_empty(self) -> None:
        """File.hash_digest cannot be empty."""
        with self.assertRaisesRegex(IntegrityError, "db_file_sha256_not_empty"):
            File.objects.get_or_create(hash_digest=b"", size=5)


class FileStoreTests(TestCase):
    """Tests for the FileStore class."""

    @override
    def setUp(self) -> None:
        """Set up FileStore to be used in the tests."""
        super().setUp()
        self.backend = FileStore.BackendChoices.LOCAL
        self.file_store_name = "nas-01"

        self.file_store = FileStore.objects.create(
            name=self.file_store_name,
            backend=self.backend,
        )

    def test_default_values(self) -> None:
        """Test default values."""
        file_store = FileStore.objects.create(
            name="nas-02", backend=FileStore.BackendChoices.LOCAL
        )
        self.assertEqual(file_store.configuration, {})
        file_store.clean_fields()
        file_store.save()

    def test_backend_choice(self) -> None:
        """Assert FileStore.BackendChoices is correctly accessed."""
        self.assertEqual(self.file_store.backend, self.backend)

    def test_configuration_validation_error(self) -> None:
        """`ValidationError` is raised if configuration is invalid."""
        file_store = FileStore.objects.create(
            name="test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration=[],
        )
        with self.assertRaises(ValidationError) as raised:
            file_store.full_clean()
        self.assertEqual(
            raised.exception.message_dict,
            {"configuration": ["configuration must be a dictionary"]},
        )

        file_store.configuration = {"nonexistent": ""}
        with self.assertRaises(ValidationError) as raised:
            file_store.full_clean()
        messages = raised.exception.message_dict
        self.assertCountEqual(messages.keys(), ["configuration"])
        self.assertRegex(
            messages["configuration"][0],
            r"(?s)invalid file store configuration:.*"
            r"Extra inputs are not permitted",
        )

    def test_files_through(self) -> None:
        """FileStore.files() return the expected file."""
        # Create a new file
        file = self.playground.create_file()

        # Add the file in the store
        FileInStore.objects.create(file=file, store=self.file_store, data={})

        # Assert that self.file_store through model return the expected file
        self.assertEqual(self.file_store.files.first(), file)

    def test_total_size(self) -> None:
        """Adding/removing files to/from the store maintains `total_size`."""
        files = [
            self.playground.create_file(contents=b"\0" * length)
            for length in (0, 4, 10, 1024, 512)
        ]

        def assert_total_size(
            expected_size_changes: Sequence[int], expected_total_size: int
        ) -> None:
            self.assertQuerySetEqual(
                FileStoreJournal.objects.filter(store=self.file_store)
                .order_by("id")
                .values_list("size_change", flat=True),
                expected_size_changes,
            )
            self.assertEqual(
                (
                    FileStore.objects.annotate_current_total_size()
                    .values_list("current_total_size", flat=True)
                    .get(pk=self.file_store.pk)
                ),
                expected_total_size,
            )

        expected_size_changes: list[int] = []
        expected_total_size = 0
        assert_total_size(expected_size_changes, expected_total_size)
        for file in files:
            FileInStore.objects.create(file=file, store=self.file_store)
            expected_size_changes.append(file.size)
            expected_total_size += file.size
            assert_total_size(expected_size_changes, expected_total_size)
        for file in files:
            FileInStore.objects.filter(
                file=file, store=self.file_store
            ).delete()
            expected_size_changes.append(-file.size)
            expected_total_size -= file.size
            assert_total_size(expected_size_changes, expected_total_size)

    def test_max_size_constraint(self) -> None:
        """If both are set, `soft_max_size` must be <= `max_size`."""
        for soft_max_size, max_size, ok in (
            (None, None, True),
            (1024, None, True),
            (None, 1024, True),
            (1024, 1024, True),
            (1024, 2048, True),
            (2048, 1024, False),
        ):
            with transaction.atomic():
                self.file_store.soft_max_size = soft_max_size
                self.file_store.max_size = max_size
                if ok:
                    self.file_store.save()
                else:
                    with self.assertRaisesRegex(
                        IntegrityError, "db_filestore_consistent_max_sizes"
                    ):
                        self.file_store.save()

    def test_get_backend_object(self) -> None:
        """get_backend_object instantiates the right backend."""
        fs_local = FileStore(
            name="local",
            backend=FileStore.BackendChoices.LOCAL,
            configuration={"base_directory": "/tmp"},
        )
        fs_memory = FileStore(
            name="memory",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "test"},
        )
        fs_external = FileStore(
            name="external",
            backend=FileStore.BackendChoices.EXTERNAL_DEBIAN_SUITE,
            configuration={
                "archive_root_url": "https://deb.debian.org/debian",
                "suite": "bookworm",
                "components": ["main"],
            },
        )

        self.assertIsInstance(fs_local.get_backend_object(), LocalFileBackend)
        self.assertIsInstance(fs_memory.get_backend_object(), MemoryFileBackend)
        self.assertIsInstance(
            fs_external.get_backend_object(), ExternalDebianSuiteFileBackend
        )

    def test_default(self) -> None:
        """default() returns FileStore with name=="Default"."""
        self.assertEqual(FileStore.default().name, "Default")

    def test_str(self) -> None:
        """__str__() return the correct information."""
        self.assertEqual(
            self.file_store.__str__(),
            f"Id: {self.file_store.id} "
            f"Name: {self.file_store.name} "
            f"Backend: {self.file_store.backend}",
        )


class FileInStoreTests(TestCase):
    """Tests for the FileInStore class."""

    @override
    def setUp(self) -> None:
        """Set up FileInStore to be used in the tests."""
        super().setUp()
        self.file_store_name = "nas-01"
        self.file_store = FileStore.objects.create(
            name="nas-01",
            backend=FileStore.BackendChoices.LOCAL,
            configuration={},
        )

        file_contents = b"test"

        self.file = self.playground.create_file(file_contents)
        self.file_in_store = FileInStore(
            store=self.file_store,
            file=self.file,
        )

    def test_default_values(self) -> None:
        """Test default values."""
        # Delete all FileInStore to create a new one, in this test,
        # reusing self.file_store and self.file
        FileInStore.objects.all().delete()
        file_in_store = FileInStore(store=self.file_store, file=self.file)

        self.assertEqual(file_in_store.data, {})
        file_in_store.clean_fields()
        file_in_store.save()

    def test_str(self) -> None:
        """__str__() return the correct string."""
        self.assertEqual(
            self.file_in_store.__str__(),
            f"Id: {self.file_in_store.id} "
            f"Store: {self.file_in_store.store.name} "
            f"File: {self.file_in_store.file.hash_digest.hex()}",
        )


class FileInStoreTransactionTests(TransactionTestCase):
    """Test transactional behaviour of FileInStore."""

    def test_parallel_add(self) -> None:
        """Two transactions that each add a file do not block each other."""
        file_store = FileStore.objects.create(
            name="memory",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "memory"},
        )
        files = [
            self.playground.create_file(contents=b"\0" * length)
            for length in (1, 2)
        ]

        with transaction.atomic():
            file_store.files.add(files[0])
            thread = RunInParallelTransaction(
                lambda: file_store.files.add(files[1])
            )
            thread.run_and_wait()

        self.assertQuerySetEqual(file_store.files.all(), files, ordered=False)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Template tag library to render Debusine UI elements."""

import abc
import datetime as dt
import logging
from collections.abc import Callable
from typing import Any, Never, override

from django import template
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.models import AnonymousUser
from django.contrib.messages.storage.base import Message
from django.db.models import Model
from django.template import Context, Node, NodeList, TemplateSyntaxError
from django.template.base import FilterExpression, Parser, Token
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.html import conditional_escape, format_html
from django.utils.safestring import SafeString
from django.utils.timesince import timesince
from more_itertools import peekable

import debusine.web.views.view_utils
from debusine.db.context import context
from debusine.db.permissioncontext import PermissionContext, PermissionUser
from debusine.server.middlewares.scopes import (
    TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL,
)
from debusine.server.scopes import urlconf_scope
from debusine.web.helps import HELPS
from debusine.web.icons import Icons
from debusine.web.views.base import Widget
from debusine.web.views.places import Place
from debusine.web.views.ui.base import UI

logger = logging.getLogger("debusine.web")

register = template.Library()


class DebusineNodeRenderError(Exception):
    """Signal an error during rendering of a DebusineNode."""

    def __init__(
        self,
        *,
        devel_message: str,
        user_message: str | None = None,
        exception: Exception | None = None,
    ) -> None:
        """
        Store exception arguments.

        :param devel_message: message targeted at developers, shown in testing
          and logs but not to the user, to prevent leaking sensitive information
        :param user_message: message shown to the user. If not provided a
          default message is used.
        :param exception: exception (if any) that caused the error
        """
        self.devel_message = devel_message
        self.user_message = user_message or "template rendering error"
        self.exception = exception


class DebusineNode(Node, abc.ABC):
    """Django Node element with added helper functions."""

    def _render_error(
        self,
        *,
        user_message: str,
        devel_message: str,
        exception: Exception | None = None,
    ) -> SafeString:
        """
        Render a description for a rendering error.

        Depending on settings, it can either raise an exception or render a
        description suitable for template output.
        """
        if settings.DEBUG or getattr(settings, "TEST_MODE", False):
            if exception is not None:
                exception.add_note(devel_message)
                raise exception
            else:
                raise ValueError(devel_message)

        # When running in production, avoid leaking possibly sensitive error
        # information while providing enough information for a potential bug
        # report to locate the stack trace in the logs
        logger.warning(
            "%s rendering error: %s",
            self.__class__.__name__,
            devel_message,
            exc_info=exception,
        )
        return format_html(
            "<span data-role='debusine-template-error'"
            " class='bg-danger text-white'>{ts} UTC: {message}</span>",
            ts=dt.datetime.now(dt.UTC).isoformat(),
            message=user_message,
        )

    def resolve_filter(
        self,
        context: Context,
        value: FilterExpression,
        value_name: str,
        fail_if_none: bool = True,
    ) -> Any:
        """
        Resolve a filter expression.

        If ``fail_if_none`` is False, returns None if the filter expression
        failed. Raises DebusineNodeRenderError on other kind of failures.
        """
        try:
            # ignore_failures seems counterintuitive to me.
            # If False, then django will *handle* failures, generating a value
            # using `string_if_invalid` configured in the template `Engine`.
            # If True, then django will *not* handle failures, and use None for
            # an undefined variable, which is something that we can look for.
            res = value.resolve(context, ignore_failures=True)
        except Exception as e:
            raise DebusineNodeRenderError(
                user_message="template argument malformed",
                devel_message=f"Invalid {value_name} argument: {str(value)!r}",
                exception=e,
            )

        if res is None and fail_if_none:
            raise DebusineNodeRenderError(
                user_message="template argument lookup failed",
                devel_message=f"{value_name} argument {str(value)!r}"
                " failed to resolve",
            )

        return res

    @abc.abstractmethod
    def render_checked(self, context: Context) -> str | SafeString:
        """
        Render the template node.

        Exceptions raised in this method are caught and rendered and logged as
        appropriate.
        """

    @override
    def render(self, context: Context) -> str | SafeString:
        try:
            return self.render_checked(context)
        except DebusineNodeRenderError as de:
            return self._render_error(
                devel_message=de.devel_message,
                user_message=de.user_message,
                exception=de.exception,
            )
        except Exception as e:
            return self._render_error(
                user_message="template rendering error",
                devel_message="unhandled exception in render_checked",
                exception=e,
            )


class WidgetNode(DebusineNode):
    """Template node implementing the {% widget %} tag."""

    def __init__(self, value: FilterExpression) -> None:
        """Store node arguments."""
        self.value = value

    @override
    def render_checked(self, context: Context) -> str | SafeString:
        # Validate the value as a widget
        match value := self.resolve_filter(context, self.value, "widget"):
            case str():
                if context.autoescape:
                    return conditional_escape(value)
                else:
                    return value
            case Widget():
                try:
                    return value.render(context)
                except Exception as e:
                    raise DebusineNodeRenderError(
                        user_message=(
                            f"{value.__class__.__name__} failed to render"
                        ),
                        devel_message=(
                            f"Widget {self.value!r} ({value!r})"
                            " failed to render"
                        ),
                        exception=e,
                    )
            case _:
                raise DebusineNodeRenderError(
                    user_message="invalid widget type",
                    devel_message=(
                        f"widget {self.value!r} {value!r} has invalid type"
                    ),
                )


@register.tag
def widget(parser: Parser, token: Token) -> WidgetNode:
    """Parser for the {% widget %} tag."""
    bits = token.split_contents()
    if len(bits) == 2:
        _, value = bits
    else:
        raise TemplateSyntaxError("{% widget %} requires exactly one argument")
    return WidgetNode(parser.compile_filter(value))


class WithScopeNode(Node):
    """Template node implementing the {% withscope %} tag."""

    def __init__(self, value: FilterExpression, nodelist: NodeList) -> None:
        """Store node arguments."""
        self.value = value
        self.nodelist = nodelist

    @override
    def render(self, context: Context) -> SafeString:
        from debusine.db.context import context as appcontext
        from debusine.db.models import Scope

        # Resolve the argument token to a value
        value: str | Scope = self.value.resolve(context)

        # Resolve a scope name to a Scope object
        match value:
            case str():
                try:
                    value = Scope.objects.get(name=value)
                except Scope.DoesNotExist:
                    return self.nodelist.render(context)
            case Scope():
                pass
            case _:
                return self.nodelist.render(context)

        pc = appcontext.pc
        with urlconf_scope(value.name), appcontext.local():
            # Render the contained template using a different current scope
            appcontext.reset()
            # Note: this triggers an extra database query, since set_user needs
            # to go and load what roles the user has on the scope.
            # TODO: if this is a problem in big querysets, the same can be done
            #       with Scope as with Workspace.objects.with_role_annotations
            appcontext.set_permission_context_scoped(value, pc)
            # Workspace is not preserved, since the scope changed
            # worker_token is not preserved since it makes no sense for
            #              templates
            # TODO: permission_checks_disabled is not currently preserved as
            #       there is no use case to justify it
            return self.nodelist.render(context)


@register.tag
def withscope(parser: Parser, token: Token) -> WithScopeNode:
    """Parser for the {% withscope %} tag."""
    bits = token.split_contents()
    if len(bits) == 2:
        _, value = bits
    else:
        raise TemplateSyntaxError("withscope requires exactly one argument")

    nodelist = parser.parse(("endwithscope",))
    parser.delete_first_token()

    return WithScopeNode(parser.compile_filter(value), nodelist)


@register.filter
def has_perm(resource: Model, name: str) -> bool:
    """Check a permission predicate."""
    predicate = getattr(resource, name)
    result = predicate(context.pc)
    assert isinstance(result, bool)
    return result


@register.filter
def roles(resource: Model, user: PermissionUser | None = None) -> list[str]:
    """Return the list of roles the user has on the resource."""
    if (get_roles := getattr(resource, "get_roles", None)) is None:
        return []
    pc: PermissionContext
    match user:
        case None:
            pc = context.pc
        case AnonymousUser():
            pc = PermissionContext.for_ui_anonymous()
        case _:
            pc = PermissionContext.for_ui_user(user)
    return sorted(get_roles(pc))


@register.filter
def as_yaml(value: Any, flags: str = "") -> SafeString:
    """Return syntax-highlighted HTML for value serialized as YAML."""
    sort_keys = "unsorted" not in flags.split(",")
    return debusine.web.views.view_utils.highlight_as_yaml(
        value, sort_keys=sort_keys
    )


@register.simple_tag
def icon(name: str) -> str:
    """Lookup an icon by name."""
    return "bi-" + getattr(Icons, name.upper(), "square-fill")


@register.filter(name="sorted")
def sorted_(value: Any) -> Any:
    """Sort a sequence of values."""
    return sorted(value)


@register.filter
def is_list_like(value: Any) -> bool:
    """Check if a value is list-like."""
    return isinstance(value, (list, tuple))


@register.filter
def is_dict_like(value: Any) -> bool:
    """Check if a value is dict-like."""
    return isinstance(value, dict)


# https://code.djangoproject.com/ticket/12486
@register.filter
def lookup[KT, VT](d: dict[KT, VT], key: KT) -> VT:
    """Get a value from a dictionary."""
    return d[key]


@register.simple_tag(name="help")
def _help(name: str) -> str:
    """Return HTML with a "?" icon and a popover with the html."""
    return render_to_string("web/_help.html", HELPS[name]._asdict())


class UINode(DebusineNode):
    """Template node implementing the {% ui %} tag."""

    def __init__(self, objvar: FilterExpression, asvar: str) -> None:
        """Store node arguments."""
        self.objvar = objvar
        self.asvar = asvar

    @override
    def render_checked(self, context: Context) -> SafeString:
        """Look up the helper and store it in the context."""
        context[self.asvar] = None
        match instance := self.resolve_filter(context, self.objvar, "resource"):
            case Model():
                try:
                    context[self.asvar] = UI.for_instance(
                        context["request"], instance
                    )
                except Exception as e:
                    raise DebusineNodeRenderError(
                        user_message="template argument lookup failed",
                        devel_message=f"ui helper lookup for {instance!r}"
                        " failed",
                        exception=e,
                    )

            case _:
                raise DebusineNodeRenderError(
                    user_message="template argument lookup failed",
                    devel_message=f"ui model instance {self.objvar!r}"
                    f" {instance!r} has invalid type",
                )
        return SafeString()


@register.tag
def ui(parser: Parser, token: Token) -> UINode:
    """
    Lookup the UI helper for a model object.

    Usage::

        {% ui file as fileui %}

        ...use fileui at will...
    """
    bits = token.split_contents()
    asvar = None
    if len(bits) == 4 and bits[-2] == "as":
        objvar = bits[1]
        asvar = bits[3]
    else:
        raise TemplateSyntaxError(
            "'ui' statement syntax is {% ui [object] as [variable] %}"
        )
    return UINode(parser.compile_filter(objvar), asvar)


class PlaceNode(DebusineNode):
    """Template node implementing the {% place %} tag."""

    def __init__(
        self,
        objvar: FilterExpression,
        place_type: str | None,
        widget_type: str,
        kwargs: list[tuple[str, FilterExpression]],
    ) -> None:
        """Store node arguments."""
        self.objvar = objvar
        self.place_type = place_type
        self.widget_type = widget_type
        self.kwargs = kwargs

    def _resolve_place(self, context: Context) -> Place:
        """Resolve self.objvar into a Place."""
        place_method_name = (
            "place" if self.place_type is None else f"place_{self.place_type}"
        )

        instance = self.resolve_filter(context, self.objvar, "object")

        if isinstance(instance, Model):
            try:
                instance = UI.for_instance(context["request"], instance)
            except Exception as e:
                raise DebusineNodeRenderError(
                    user_message="ui lookup failed",
                    devel_message=f"ui helper lookup for {self.objvar!r}"
                    " failed",
                    exception=e,
                )

        if isinstance(instance, UI):
            try:
                instance = getattr(instance, place_method_name)
            except Exception as e:
                raise DebusineNodeRenderError(
                    user_message="place lookup failed",
                    devel_message=f"place method {place_method_name!r}"
                    f" for {instance!r} failed",
                    exception=e,
                )

        if isinstance(instance, Place):
            return instance

        raise DebusineNodeRenderError(
            user_message="invalid object type",
            devel_message=f"object {self.objvar!r} resolved to unsupported"
            f" object type {instance.__class__.__name__} ({instance!r})",
        )

    def _resolve_kwargs(self, context: Context) -> dict[str, Any]:
        """Resolve self.kwargs into a dict of keyword arguments."""
        kwargs: dict[str, Any] = {}
        for key, value_lookup in self.kwargs:
            kwargs[key] = self.resolve_filter(
                context, value_lookup, "kwarg value"
            )
        return kwargs

    @override
    def render_checked(self, context: Context) -> str:
        place = self._resolve_place(context)

        # Resolve render method
        renderer: Callable[..., str] | None = getattr(
            place, self.widget_type, None
        )
        if renderer is None:
            raise DebusineNodeRenderError(
                user_message="invalid place render type",
                devel_message=f"place {place!r} does not have"
                f" an {self.widget_type} method",
            )

        # Resolve kwargs
        kwargs = self._resolve_kwargs(context)

        # Call the render method
        try:
            return renderer(**kwargs)
        except Exception as e:
            raise DebusineNodeRenderError(
                user_message="place rendering failed",
                devel_message=f"place {place!r} {self.widget_type}"
                " method failed",
                exception=e,
            )


def _place_template_syntax_error(message: str) -> Never:
    """Raise TemplateSyntaxError for ``{% place %}`` tags."""
    raise TemplateSyntaxError(
        message + ". Usage: {% place {object} [place_type] as_{widget_type}"
        " [key=value, ...] %}"
    )


@register.tag
def place(parser: Parser, token: Token) -> PlaceNode:
    """
    Render Place elements.

    Usage::

        {% place {Model|UI|Place} [type] as_[widget] [key=value, ...] %}

    The first argument is used to obtain a Place instance:

    * if it's a ``Place`` instance, it is used
    * if it's a UI helper instance, ``place`` or ``place_{type}`` is invoked
    * if it's a ``Model`` instance, its UI helper is instantiated and used to
      get a ``Place`` as in the previous point

    Then the ``as_[widget]`` method is called on the ``Place`` instance to
    render it. Any trailing ``key=value`` assignments are passed to the
    ``as_[widget]`` method.
    """
    bits = token.split_contents()

    args = peekable(bits)
    next(args)  # discard "place"

    # Get resource/ui helper/place
    if not args:
        _place_template_syntax_error("object not provided")
    obj: str = next(args)

    # Get optional place type
    place_type: str | None = None
    if not args:
        _place_template_syntax_error("as_[widget] not provided")
    if not args.peek().startswith("as_"):
        place_type = next(args)

    # Get widget type
    if not args:
        _place_template_syntax_error("as_[widget] not provided")
    widget_type: str = next(args)
    if not widget_type.startswith("as_"):
        _place_template_syntax_error("widget type does not begin with `as_`")

    kwargs: list[tuple[str, FilterExpression]] = []
    for kwarg in args:
        if "=" not in kwarg:
            _place_template_syntax_error("missing = in key=value assignment")
        key, arg = kwarg.split("=", 1)
        if not key:
            _place_template_syntax_error("empty key in key=value assignment")
        kwargs.append((key, parser.compile_filter(arg)))

    return PlaceNode(
        parser.compile_filter(obj), place_type, widget_type, kwargs
    )


@register.filter
def message_toast_color_class(message: Message) -> str:
    """Get the bootstrap color class to use for showing Django messages."""
    match message.level:
        case messages.SUCCESS:
            return "success"
        case messages.WARNING:
            return "warning"
        case messages.ERROR:
            return "danger"
        case _:
            return "light"


@register.filter
def message_toast_header_icon(message: Message) -> str:
    """Get toast header classes to use for showing Django messages."""
    match message.level:
        case messages.WARNING | messages.ERROR:
            return "bi bi-exclamation-triangle"
        case _:
            return "bi bi-info-circle"


@register.filter
def worker_last_seen(value: dt.datetime | None) -> str:
    """Return human-readable text for a worker's last-seen timestamp."""
    if value is None:
        return "Never"

    if timezone.now() - value <= TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL:
        minutes = int(TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL.total_seconds() / 60)
        return f"Less than {minutes} minutes ago"

    return f"{timesince(value)} ago"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data used to display help."""

from typing import NamedTuple

from django.utils.safestring import SafeString

DOCS_BASE_URL = "https://freexian-team.pages.debian.net/debusine"


class Help(NamedTuple):
    """Represents help metadata."""

    title: str
    summary: SafeString
    link: str


HELPS: dict[str, Help] = {
    "dependencies": Help(
        title="Dependencies",
        summary=SafeString(
            "Work requests that must complete "
            "to unblock this one and let it run."
        ),
        link=f"{DOCS_BASE_URL}/explanation/"
        "workflow-orchestration.html#dependencies",
    ),
    "reverse_dependencies": Help(
        title="Reverse dependencies",
        summary=SafeString(
            "Work requests that require this one to complete "
            "to unblock and let them run."
        ),
        link=f"{DOCS_BASE_URL}/explanation/"
        "workflow-orchestration.html#dependencies",
    ),
    "expiration_delay": Help(
        title="Expiration delay",
        summary=SafeString(
            "Days after which objects such as artifacts or "
            "work requests are removed"
        ),
        link=f"{DOCS_BASE_URL}/explanation/expiration-of-data.html",
    ),
    "workspace_inheritance": Help(
        title="Workspace inheritance",
        summary=SafeString(
            "Workspaces can inherit from one another in order to share"
            " collections and artifacts when required."
        ),
        link=f"{DOCS_BASE_URL}/reference/workspaces/inheritance.html",
    ),
}



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for tar related classes."""

import errno
import io
import tarfile
from pathlib import Path
from unittest import mock

from debusine.server.tar import TarArtifact
from debusine.test.django import TestCase


class TarArtifactTests(TestCase):
    """Tests for TarArtifact class."""

    playground_memory_file_store = False

    def test_tar_small_chunks(self) -> None:
        """Assert that TarArtifact return data in small chunks."""
        file_paths = ["README"]

        artifact, _ = self.playground.create_artifact(
            file_paths, create_files=True
        )

        tar_artifact = TarArtifact(artifact=artifact)

        tar_artifact._max_chunk_size = max_chunk_size = 5

        for chunk in tar_artifact:
            self.assertLessEqual(len(chunk), max_chunk_size)

        self.assertTrue(tar_artifact._tar_file.closed)

    def test_tar_subdirectory(self) -> None:
        """Test creating a tar for an artifact, only files from a subdir."""
        file_paths = ["README", "doc/README", "doc/README2", "src/main.c"]

        artifact, _ = self.playground.create_artifact(
            file_paths, create_files=True
        )

        tar_artifact = TarArtifact(artifact=artifact, subdirectory="doc/")

        result = io.BytesIO(b"".join(tar_artifact))
        result.seek(0)

        with tarfile.open(fileobj=result, mode="r:gz") as tar_artifact_result:
            self.assertEqual(
                tar_artifact_result.getnames(), ["doc/README", "doc/README2"]
            )

        self.assertTrue(tar_artifact._tar_file.closed)

    def test_tar(self) -> None:
        """Test creating a tar for an artifact."""
        file_paths = ["README", "README2", "doc/README3"]

        artifact, files_contents = self.playground.create_artifact(
            file_paths, create_files=True
        )

        tar_artifact = TarArtifact(artifact=artifact)

        result = io.BytesIO()

        # Verify that TarArtifact generate multiple chunks (at least
        # one per file) and that the file in disk decreased it size
        # (because the returned data is removed from the file)
        loops = 0

        for chunk in tar_artifact:
            result.write(chunk)
            loops += 1

        # Assert that TarArtifact generated at least as many
        # chunks as number of files
        self.assertGreaterEqual(loops, len(file_paths))

        result.seek(0)

        with tarfile.open(fileobj=result, mode="r:gz") as tar_file:
            # Assert file's contents are correct
            for path in file_paths:
                file = tar_file.extractfile(path)
                assert file is not None
                self.assertEqual(file.read(), files_contents[path])

            # Assert file names are ordered
            self.assertEqual(
                tar_file.getnames(),
                list(
                    artifact.fileinartifact_set.values_list(
                        "path", flat=True
                    ).order_by("path")
                ),
            )

            # Assert mtime of each file is artifact.created_at
            for member in tar_file.getmembers():
                self.assertEqual(member.mtime, artifact.created_at.timestamp())

        # Assert TarArtifact deleted the temporary directory when it finished
        # iterating
        self.assertFalse(Path(tar_artifact._temp_directory).exists())

        self.assertTrue(tar_artifact._tar_file.closed)

    def test_tar_excludes_incomplete(self) -> None:
        """Creating a tar for an artifact excludes incomplete files."""
        file_paths = ["README", "README2", "doc/README3"]

        artifact, _ = self.playground.create_artifact(
            file_paths, create_files=True
        )
        artifact.fileinartifact_set.filter(path="README").update(complete=False)

        tar_artifact = TarArtifact(artifact=artifact)

        result = io.BytesIO(b"".join(tar_artifact))
        result.seek(0)

        with tarfile.open(fileobj=result, mode="r:gz") as tar_artifact_result:
            self.assertCountEqual(
                tar_artifact_result.getnames(), ["README2", "doc/README3"]
            )

        self.assertTrue(tar_artifact._tar_file.closed)

    def test_next_raise_stop_iteration_file_not_found(self) -> None:
        """Cannot read a file: exc logged, StopIteration raised."""
        file_paths = ["README"]

        artifact, _ = self.playground.create_artifact(
            file_paths, create_files=True
        )

        fileobj = artifact.fileinartifact_set.all()[0].file
        file_backend = artifact.workspace.scope.download_file_backend(fileobj)

        entry = file_backend.get_entry(fileobj)
        entry.remove()

        tar_artifact = TarArtifact(artifact=artifact)

        with (
            self.assertLogsContains(
                str(entry.get_local_path()), expected_count=2
            ),
            self.assertRaises(StopIteration),
        ):
            tar_artifact.__next__()

        self.assertTrue(tar_artifact._tar_file.closed)

    def test_resource_cleanup_on_error(self) -> None:
        file_paths = ["README"]
        artifact, _ = self.playground.create_artifact(
            file_paths, create_files=True
        )

        tar_artifact = TarArtifact(artifact)
        self.assertFalse(tar_artifact._tar_closed)
        self.assertFalse(tar_artifact._tar_file.closed)

        # Simulate a read error to close the streaming prematurely
        with mock.patch.object(
            tar_artifact._tar_file,
            "read",
            side_effect=OSError(errno.EINTR, "expected error"),
        ):
            list(tar_artifact)

        self.assertTrue(tar_artifact._closed)
        self.assertTrue(tar_artifact._tar_closed)
        self.assertTrue(tar_artifact._tar_file.closed)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for common Django utility code."""

import io
import sys
from collections.abc import Generator
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from typing import Any, TextIO

from django.core.management import call_command as django_call_command


def call_command(
    cmd: str,
    *args: str,
    verbosity: int = 0,
    stdin: TextIO = sys.stdin,
    **kwargs: Any,
) -> tuple[str, str, int | str]:
    """
    Call command cmd with arguments.

    Capture SystemExit's code.

    Return tuple with:
    - stdout's of the command
    - stderr's of the command
    - exit_code of the command or the string "command did not raise SystemExit \
      or call sys.exit"
    """
    stdout = io.StringIO()
    stderr = io.StringIO()
    exit_code: int | str = "command did not raise SystemExit or call sys.exit"

    try:
        with (
            redirect_stderr(stderr),
            redirect_stdout(stdout),
            _redirect_stdin(stdin),
        ):
            django_call_command(cmd, *args, **kwargs, verbosity=verbosity)
    except SystemExit as exc:
        exit_code = 0 if exc.code is None else exc.code

    return stdout.getvalue(), stderr.getvalue(), exit_code


@contextmanager
def _redirect_stdin(new_stdin: TextIO) -> Generator[None]:
    """Context manager to temporarily redirect stdin."""
    old_stdin = sys.stdin
    sys.stdin = new_stdin
    try:
        yield
    finally:
        sys.stdin = old_stdin



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for task status view."""

from typing import override

from django.template.loader import get_template
from django.test import RequestFactory, override_settings
from django.urls import reverse

import debusine.worker.tags as wtags
from debusine.db.context import context
from debusine.db.models import WorkRequest
from debusine.db.playground import scenarios
from debusine.server.status import TaskQueueSummary, WorkerSummary
from debusine.tasks.models import WorkerType
from debusine.test.django import TestCase
from debusine.web.views.places import IndexPlace
from debusine.web.views.task_status import TaskStatusView
from debusine.web.views.tests.utils import ViewTestMixin


class TaskStatusViewTests(ViewTestMixin, TestCase):
    """Tests for TaskStatusView."""

    scenario = scenarios.DefaultContext(set_current=True)

    @override
    def setUp(self) -> None:
        super().setUp()
        self.factory = RequestFactory()

    def test_get_place(self) -> None:
        view = TaskStatusView()
        view.request = RequestFactory().get(reverse("task-status"))
        place = view.get_place()
        self.assertIsInstance(place, IndexPlace)
        self.assertEqual(place.title, "Task queue")
        self.assertEqual(place.url, reverse("task-status"))
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)

    def test_template_no_work_requests(self) -> None:
        """Test template display the "No work requests..." message."""
        response = self.client.get(reverse("task-status"))

        self.assertContains(
            response, "<p>There are no pending work requests.</p>", html=True
        )

    def test_template_simple(self) -> None:
        """Test template row for arch-None without any tasks."""
        self.playground.create_worker(
            worker_type=WorkerType.EXTERNAL,
            extra_dynamic_metadata={
                "system:native_architecture": "amd64",
                "system:architectures": ["amd64", "i386"],
            },
        ).mark_connected()
        response = self.client.get(reverse("task-status"))
        tree = self.assertResponseHTML(response)
        arch_table = self.assertHasElement(
            tree, "//table[@id='task-arch-queue']"
        )
        self.assertHasElement(arch_table.tbody, "tr[@id='arch-_none_']")

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_template(self) -> None:
        """Test template contains expected output."""
        source = self.playground.create_minimal_source_package_artifact()
        environment = self.playground.create_debian_environment().artifact
        assert environment is not None

        worker = self.playground.create_worker(
            worker_type=WorkerType.EXTERNAL,
            extra_dynamic_metadata={
                "system:native_architecture": "amd64",
                "system:architectures": ["amd64", "i386"],
            },
        )
        worker.mark_connected()

        # One pending amd64 task with a worker assigned, but not yet running
        task_amd64 = self.playground.create_sbuild_work_request(
            source=source, environment=environment, architecture="amd64"
        )
        self.playground.advance_work_request(task_amd64, worker=worker)

        # Create a task that is (oddly) running without an assigned worker.
        unassigned_wr = self.playground.create_sbuild_work_request(
            source=source, environment=environment, architecture="i386"
        )
        self.playground.advance_work_request(unassigned_wr, mark_pending=True)
        unassigned_wr.status = WorkRequest.Statuses.RUNNING
        unassigned_wr.save()

        # Noop task currently uses the fallback architecture
        self.playground.create_worker_task(task_name="noop", mark_pending=True)

        # Task that successfully transitioned to pending but now fails to
        # validate
        corrupted_wr = self.playground.create_worker_task(
            task_name="sbuild",
            mark_pending=True,
            validate=False,
            use_data=False,
        )
        corrupted_wr.scheduler_tags_required = [
            wtags.BUILD_ARCH_PREFIX + "amd64"
        ]
        corrupted_wr.save()

        response = self.client.get(reverse("task-status"))
        tree = self.assertResponseHTML(response)

        type_table = self.assertHasElement(tree, "//table[@id='task-queue']")
        row = self.assertHasElement(
            type_table.tbody, "tr[@id='worker_type-external']"
        )
        # Columns: type, registered, connected, busy, idle, running, pending
        self.assertHTMLContentsEquivalent(row.td[0], "<td>External</td>")
        self.assertHTMLContentsEquivalent(row.td[1], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[2], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[3], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[4], "<td>0</td>")
        self.assertHTMLContentsEquivalent(row.td[5], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[6], "<td>3</td>")

        arch_table = self.assertHasElement(
            tree, "//table[@id='task-arch-queue']"
        )

        # First row: architecture amd64, 1 connected worker, 2 work requests
        row = self.assertHasElement(arch_table.tbody, "tr[@id='arch-amd64']")
        self.assertHTMLContentsEquivalent(row.td[0], '<td>amd64</td>')
        # Columns: registered, connected, busy, idle, running, pending
        self.assertHTMLContentsEquivalent(row.td[1], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[2], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[3], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[4], "<td>0</td>")
        self.assertHTMLContentsEquivalent(row.td[5], "<td>0</td>")
        self.assertHTMLContentsEquivalent(row.td[6], "<td>2</td>")

        # Second row: architecture i386, 1 connected worker, 0 work requests
        row = arch_table.tbody.xpath("tr[@id='arch-i386']")[0]
        self.assertHTMLContentsEquivalent(row.td[0], '<td>i386</td>')
        self.assertHTMLContentsEquivalent(row.td[1], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[2], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[3], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[4], "<td>0</td>")
        self.assertHTMLContentsEquivalent(row.td[5], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[6], "<td>0</td>")

        # Third row: arch unspecified, 1 connected worker, 2 work requests
        row = arch_table.tbody.xpath("tr[@id='arch-_none_']")[0]
        self.assertHTMLContentsEquivalent(row.td[0], "<td>Not Specified</td>")
        self.assertHTMLContentsEquivalent(row.td[1], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[2], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[3], "<td>1</td>")
        self.assertHTMLContentsEquivalent(row.td[4], "<td>0</td>")
        self.assertHTMLContentsEquivalent(row.td[5], "<td>0</td>")
        self.assertHTMLContentsEquivalent(row.td[6], "<td>1</td>")

        # Three rows for two architectures: amd64 and i386, then None
        self.assertEqual(len(arch_table.tbody.tr), 3)

    def test_task_queue_sorted_in_template(self) -> None:
        """task_queue is sorted by architecture in the template."""
        worker_status = {
            WorkerType.EXTERNAL: {
                "worker_tasks": TaskQueueSummary(),
                "workers": WorkerSummary(),
            },
        }
        # Create an unsorted arch_status dictionary
        arch_status = {
            "i386": {
                "external_workers_arch": WorkerSummary(connected=3),
                "worker_tasks_arch": TaskQueueSummary(pending=5),
            },
            "amd64": {
                "external_workers_arch": WorkerSummary(connected=2),
                "worker_tasks_arch": TaskQueueSummary(pending=3),
            },
        }
        template = get_template("web/task_status.html")
        # FIXME: template.render does not use context processors, so we need to
        # explicitly re-add here what the template expects. This may get more
        # complicated in the future, as the base template may get to rely more
        # on default context
        rendered_html = template.render(
            {
                "arch_status": arch_status,
                "worker_status": worker_status,
                "workspace": context.workspace,
                "base_template": "web/_base_debusine.html",
            }
        )

        i386_index = rendered_html.find("i386")
        amd64_index = rendered_html.find("amd64")

        self.assertLess(amd64_index, i386_index)
